import numpy as np

import openmdao.api as om


class FlightPathEOM2D(om.ExplicitComponent):
    """
    Component containing the ODE for 2D flight.

    Computes the position and velocity equations of motion using a 2D flight path
    parameterization of states per equations 4.42 - 4.46 of _[1].

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.

    References
    ----------
    .. [1] Bryson, Arthur Earl. Dynamic optimization. Vol. 1. Prentice Hall, p.172, 1999.
    """
    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare('num_nodes', types=int)

    def setup(self):
        """
        Add inputs and outputs to this component.
        """
        nn = self.options['num_nodes']

        self.add_input(name='m',
                       val=np.ones(nn),
                       desc='aircraft mass',
                       units='kg')

        self.add_input(name='v',
                       val=np.ones(nn),
                       desc='aircraft velocity magnitude',
                       units='m/s')

        self.add_input(name='T',
                       val=np.zeros(nn),
                       desc='thrust',
                       units='N')

        self.add_input(name='alpha',
                       val=np.zeros(nn),
                       desc='angle of attack',
                       units='rad')

        self.add_input(name='L',
                       val=np.zeros(nn),
                       desc='lift force',
                       units='N')

        self.add_input(name='D',
                       val=np.zeros(nn),
                       desc='drag force',
                       units='N')

        self.add_input(name='gam',
                       val=np.zeros(nn),
                       desc='flight path angle',
                       units='rad')

        self.add_output(name='v_dot',
                        val=np.zeros(nn),
                        desc='rate of change of velocity magnitude',
                        units='m/s**2')

        self.add_output(name='gam_dot',
                        val=np.zeros(nn),
                        desc='rate of change of flight path angle',
                        units='rad/s')

        self.add_output(name='h_dot',
                        val=np.zeros(nn),
                        desc='rate of change of altitude',
                        units='m/s')

        self.add_output(name='r_dot',
                        val=np.zeros(nn),
                        desc='rate of change of range',
                        units='m/s')

        ar = np.arange(nn)

        self.declare_partials('v_dot', 'T', rows=ar, cols=ar)
        self.declare_partials('v_dot', 'D', rows=ar, cols=ar)
        self.declare_partials('v_dot', 'm', rows=ar, cols=ar)
        self.declare_partials('v_dot', 'gam', rows=ar, cols=ar)
        self.declare_partials('v_dot', 'alpha', rows=ar, cols=ar)

        self.declare_partials('gam_dot', 'T', rows=ar, cols=ar)
        self.declare_partials('gam_dot', 'L', rows=ar, cols=ar)
        self.declare_partials('gam_dot', 'm', rows=ar, cols=ar)
        self.declare_partials('gam_dot', 'gam', rows=ar, cols=ar)
        self.declare_partials('gam_dot', 'alpha', rows=ar, cols=ar)
        self.declare_partials('gam_dot', 'v', rows=ar, cols=ar)

        self.declare_partials('h_dot', 'gam', rows=ar, cols=ar)
        self.declare_partials('h_dot', 'v', rows=ar, cols=ar)

        self.declare_partials('r_dot', 'gam', rows=ar, cols=ar)
        self.declare_partials('r_dot', 'v', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        """
        Compute ODE outputs.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        g = 9.80665
        m = inputs['m']
        v = inputs['v']
        T = inputs['T']
        L = inputs['L']
        D = inputs['D']
        gam = inputs['gam']
        alpha = inputs['alpha']

        calpha = np.cos(alpha)
        salpha = np.sin(alpha)

        cgam = np.cos(gam)
        sgam = np.sin(gam)

        mv = m * v

        outputs['v_dot'] = (T * calpha - D) / m - g * sgam

        outputs['gam_dot'] = (T * salpha + L) / mv - (g / v) * cgam

        outputs['h_dot'] = v * sgam

        outputs['r_dot'] = v * cgam

    def compute_partials(self, inputs, partials):
        """
        Compute sub-jacobian parts. The model is assumed to be in an unscaled state.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        partials : Jacobian
            Subjac components written to partials[output_name, input_name].
        """
        g = 9.80665
        m = inputs['m']
        v = inputs['v']
        T = inputs['T']
        L = inputs['L']
        D = inputs['D']
        gam = inputs['gam']
        alpha = inputs['alpha']

        calpha = np.cos(alpha)
        salpha = np.sin(alpha)

        cgam = np.cos(gam)
        sgam = np.sin(gam)

        mv = m * v

        partials['v_dot', 'T'] = calpha / m
        partials['v_dot', 'D'] = -1.0 / m
        partials['v_dot', 'm'] = (D - T * calpha) / (m**2)
        partials['v_dot', 'gam'] = -g * cgam
        partials['v_dot', 'alpha'] = -T * salpha / m

        partials['gam_dot', 'T'] = salpha / mv
        partials['gam_dot', 'L'] = 1.0 / mv
        partials['gam_dot', 'm'] = -(L + T * salpha) / (m * mv)
        partials['gam_dot', 'gam'] = g * sgam / v
        partials['gam_dot', 'alpha'] = T * calpha / mv
        partials['gam_dot', 'v'] = g * cgam / v**2 - (L + T * salpha) / (v * mv)

        partials['h_dot', 'gam'] = v * cgam
        partials['h_dot', 'v'] = sgam

        partials['r_dot', 'gam'] = -v * sgam
        partials['r_dot', 'v'] = cgam


if __name__ == "__main__":

    import openmdao.api as om
    p = om.Problem()
    p.model = FlightPathEOM2D(num_nodes=2)

    p.setup(force_alloc_complex=True)
    p.check_partials(method='cs')



__author__ = 'rfalck'

import os.path

import numpy as np
import openmdao.api as om
from openmdao.components.interp_util.interp import InterpND


def setup_surrogates_all(model_name='CRM'):
    data_dir = os.path.split(os.path.realpath(__file__))[0]
    raw = np.loadtxt(os.path.join(data_dir, 'data', '{0}_aero_inputs.dat'.format(model_name)))
    [CL, CD, CM] = np.loadtxt(os.path.join(data_dir, 'data',
                                           '{0}_aero_outputs.dat'.format(model_name)))

    M_num, a_num, h_num, e_num = raw[:4].astype(int)
    M_surr = raw[4:4 + M_num]
    a_surr = raw[4 + M_num:4 + M_num + a_num]
    h_surr = raw[4 + M_num + a_num:4 + M_num + a_num + h_num]
    e_surr = raw[4 + M_num + a_num + h_num:]

    interp_CL = np.zeros((M_num, a_num, h_num, e_num))
    interp_CD = np.zeros((M_num, a_num, h_num, e_num))
    interp_CM = np.zeros((M_num, a_num, h_num, e_num))

    count = 0
    for i in range(M_num):
        for j in range(a_num):
            for k in range(h_num):
                for l in range(e_num):
                    interp_CL[i][j][k][l] = CL[count]
                    interp_CD[i][j][k][l] = CD[count]
                    interp_CM[i][j][k][l] = CM[count]
                    count += 1

    interpND_CL = InterpND(method='lagrange3', points=(M_surr, a_surr, h_surr, e_surr),
                           values=interp_CL)
    interpND_CD = InterpND(method='lagrange3', points=(M_surr, a_surr, h_surr, e_surr),
                           values=interp_CD)
    interpND_CM = InterpND(method='lagrange3', points=(M_surr, a_surr, h_surr, e_surr),
                           values=interp_CM)

    nums = {
        'M': M_num,
        'a': a_num,
        'h': h_num,
        'e': e_num,
    }

    return [interpND_CL, interpND_CD, interpND_CM, nums]


class InterpNDAeroCoeffComp(om.ExplicitComponent):
    """ Compute the lift, drag, and moment coefficients of the aircraft """
    def initialize(self):
        self.options.declare('vec_size', types=int)
        self.options.declare('interpND_CL')
        self.options.declare('interpND_CD')
        self.options.declare('interpND_CM')
        self.options.declare('interp_num')

    def setup(self):
        nn = self.options['vec_size']

        # Inputs
        self.add_input(name='M', shape=(nn,), desc='Mach number', units=None)
        self.add_input(name='h', shape=(nn,), desc='Altitude', units='km')
        self.add_input(name='alpha', shape=(nn,), desc='Angle of attack', units='rad')
        self.add_input(name='eta', shape=(nn,), desc='tail rotation angle', units='rad')

        # Outputs
        self.add_output(name='CL', shape=(nn,), desc='Lift coefficient', units=None)
        self.add_output(name='CD', shape=(nn,), desc='Drag coefficient', units=None)
        self.add_output(name='CM', shape=(nn,), desc='Moment coefficient', units=None)

        # Initialization
        self.inputs = np.zeros((nn, 4))

        # interp_CL = self.options['interp_CL']
        interpND_CL = self.options['interpND_CL']
        interpND_CD = self.options['interpND_CD']
        interpND_CM = self.options['interpND_CM']

        self.interp_tup = ((0, 'CL', interpND_CL), (1, 'CD', interpND_CD), (2, 'CM', interpND_CM))

        ar = np.arange(nn)
        self.declare_partials('CL', 'M', rows=ar, cols=ar)
        self.declare_partials('CL', 'alpha', rows=ar, cols=ar)
        self.declare_partials('CL', 'h', rows=ar, cols=ar)
        self.declare_partials('CL', 'eta', rows=ar, cols=ar)

        self.declare_partials('CD', 'M', rows=ar, cols=ar)
        self.declare_partials('CD', 'alpha', rows=ar, cols=ar)
        self.declare_partials('CD', 'h', rows=ar, cols=ar)
        self.declare_partials('CD', 'eta', rows=ar, cols=ar)

        self.declare_partials('CM', 'M', rows=ar, cols=ar)
        self.declare_partials('CM', 'alpha', rows=ar, cols=ar)
        self.declare_partials('CM', 'h', rows=ar, cols=ar)
        self.declare_partials('CM', 'eta', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        self.inputs[:, 0] = inputs['M']
        self.inputs[:, 1] = np.degrees(inputs['alpha'])  # convert to deg
        self.inputs[:, 2] = inputs['h'] * 3.28e3   # convert km to ft
        self.inputs[:, 3] = np.degrees(inputs['eta'])  # convert to deg

        outputs['CL'][:] = self.options['interpND_CL'].interpolate(self.inputs)[:]
        outputs['CD'][:] = self.options['interpND_CD'].interpolate(self.inputs)[:] + 0.015
        outputs['CM'][:] = self.options['interpND_CM'].interpolate(self.inputs)[:]

    def compute_partials(self, inputs, partials):
        self.inputs[:, 0] = inputs['M']
        self.inputs[:, 1] = np.degrees(inputs['alpha'])  # convert to deg
        self.inputs[:, 2] = inputs['h'] * 3.28e3   # convert km to ft
        self.inputs[:, 3] = np.degrees(inputs['eta'])  # convert to deg

        for ind, name, interp in self.interp_tup:
            # compute_derivative
            values, derivs = interp.interpolate(self.inputs, compute_derivative=True)[:]
            partials[name, 'M'] = derivs[:, 0]
            partials[name, 'alpha'] = np.degrees(derivs[:, 1])
            partials[name, 'h'] = derivs[:, 2] * 3.28e3
            partials[name, 'eta'] = np.degrees(derivs[:, 3])



import numpy as np

import openmdao.api as om


class MassFlowRateComp(om.ExplicitComponent):
    """ Computes mass flow rate for the F4's 2 J79 engines at full throttle. """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input('thrust', shape=(nn,), desc='engine thrust', units='N')
        self.add_input('Isp', shape=(nn,), desc='engine specific impulse', units='s')

        # Outputs
        self.add_output(name='m_dot', val=np.zeros(nn),
                        desc='vehicle mass accumulation rate due to fuel consumption',
                        units='kg/s')

        # Jacobian
        ar = np.arange(nn)
        self.declare_partials(of='m_dot', wrt='thrust', rows=ar, cols=ar)
        self.declare_partials(of='m_dot', wrt='Isp', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        outputs['m_dot'] = -inputs['thrust'] / (9.80665 * inputs['Isp'])

    def compute_partials(self, inputs, partials):
        mdot = inputs['thrust'] / (9.80665 * inputs['Isp'])
        partials['m_dot', 'thrust'] = -1.0 / (9.80665 * inputs['Isp'])
        partials['m_dot', 'Isp'] = mdot / inputs['Isp']



import os
import unittest

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
plt.style.use('ggplot')

from openmdao.utils.testing_utils import use_tempdirs


@use_tempdirs
class TestBrachistochroneStaticGravity(unittest.TestCase):

    def tearDown(self):
        for filename in ['total_coloring.pkl', 'SLSQP.out', 'SNOPT_print.out', 'SNOPT_summary.out']:
            if os.path.exists(filename):
                os.remove(filename)

    def test_brachistochrone_partials(self):
        import numpy as np
        import openmdao.api as om
        from dymos.utils.testing_utils import assert_check_partials
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        num_nodes = 5

        p = om.Problem(model=om.Group())

        ivc = p.model.add_subsystem('vars', om.IndepVarComp())
        ivc.add_output('v', shape=(num_nodes,), units='m/s')
        ivc.add_output('theta', shape=(num_nodes,), units='deg')

        p.model.add_subsystem('ode', BrachistochroneODE(num_nodes=num_nodes, static_gravity=True))

        p.model.connect('vars.v', 'ode.v')
        p.model.connect('vars.theta', 'ode.theta')

        p.setup(force_alloc_complex=True)

        p.set_val('vars.v', 10*np.random.random(num_nodes))
        p.set_val('vars.theta', 10*np.random.uniform(1, 179, num_nodes))

        p.run_model()
        cpd = p.check_partials(method='cs', compact_print=True)
        assert_check_partials(cpd)

    def test_brachistochrone_static_gravity(self):
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        import matplotlib.pyplot as plt
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        #
        # Initialize the Problem and the optimization driver
        #
        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        #
        # Create a trajectory and add a phase to it
        #
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        phase = traj.add_phase('phase0',
                               dm.Phase(ode_class=BrachistochroneODE,
                                        ode_init_kwargs={'static_gravity': True},
                                        transcription=dm.GaussLobatto(num_segments=10)))

        #
        # Set the variables
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', rate_source='xdot',
                        targets=None,
                        units='m',
                        fix_initial=True, fix_final=True, solve_segments=False)

        phase.add_state('y', rate_source='ydot',
                        targets=None,
                        units='m',
                        fix_initial=True, fix_final=True, solve_segments=False)

        phase.add_state('v', rate_source='vdot',
                        targets=['v'],
                        units='m/s',
                        fix_initial=True, fix_final=False, solve_segments=False)

        phase.add_control('theta', targets=['theta'],
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', targets=['g'], static_target=True, opt=False)

        #
        # Minimize time at the end of the phase
        #
        phase.add_objective('time', loc='final', scaler=10)

        #
        # Setup the Problem
        #
        p.setup()

        #
        # Set the initial values
        # The initial time is fixed, and we set that fixed value here.
        # The optimizer is allowed to modify t_duration, but an initial guess is provided here.
        #
        p.set_val('traj.phase0.t_initial', 0.0)
        p.set_val('traj.phase0.t_duration', 2.0)

        # Guesses for states are provided at all state_input nodes.
        # We use the phase.interpolate method to linearly interpolate values onto the state input nodes.
        # Since fix_initial=True for all states and fix_final=True for x and y, the initial or final
        # values of the interpolation provided here will not be changed by the optimizer.
        p.set_val('traj.phase0.states:x', phase.interp('x', [0, 10]))
        p.set_val('traj.phase0.states:y', phase.interp('y', [10, 5]))
        p.set_val('traj.phase0.states:v', phase.interp('v', [0, 9.9]))

        # Guesses for controls are provided at all control_input node.
        # Here phase.interpolate is used to linearly interpolate values onto the control input nodes.
        p.set_val('traj.phase0.controls:theta', phase.interp('theta', [5, 100.5]))

        # Set the value for gravitational acceleration.
        p.set_val('traj.phase0.parameters:g', 9.80665)

        #
        # Solve for the optimal trajectory
        #
        dm.run_problem(p, simulate=True)

        # Test the results
        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1], 1.8016,
                          tolerance=1.0E-3)

        # Generate the explicitly simulated trajectory
        exp_out = om.CaseReader('dymos_simulation.db').get_case('final')

        # Extract the timeseries from the implicit solution and the explicit simulation
        x = p.get_val('traj.phase0.timeseries.states:x')
        y = p.get_val('traj.phase0.timeseries.states:y')
        t = p.get_val('traj.phase0.timeseries.time')
        theta = p.get_val('traj.phase0.timeseries.controls:theta')

        x_exp = exp_out.get_val('traj.phase0.timeseries.states:x')
        y_exp = exp_out.get_val('traj.phase0.timeseries.states:y')
        t_exp = exp_out.get_val('traj.phase0.timeseries.time')
        theta_exp = exp_out.get_val('traj.phase0.timeseries.controls:theta')

        fig, axes = plt.subplots(nrows=2, ncols=1)

        axes[0].plot(x, y, 'o')
        axes[0].plot(x_exp, y_exp, '-')
        axes[0].set_xlabel('x (m)')
        axes[0].set_ylabel('y (m)')

        axes[1].plot(t, theta, 'o')
        axes[1].plot(t_exp, theta_exp, '-')
        axes[1].set_xlabel('time (s)')
        axes[1].set_ylabel(r'$\theta$ (deg)')

        plt.show()


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest
from numpy.testing import assert_almost_equal

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

from openmdao.utils.general_utils import set_pyoptsparse_opt
OPT, OPTIMIZER = set_pyoptsparse_opt('SNOPT', fallback=True)


@use_tempdirs
class TestBrachistochroneVaryingOrderControlSimulation(unittest.TestCase):

    @require_pyoptsparse(optimizer='SLSQP')
    def test_gauss_lobatto(self):
        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring()

        t = dm.GaussLobatto(num_segments=20,
                            order=[3, 5]*10,
                            compressed=True)

        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=t)

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False)
        phase.add_state('y', fix_initial=True, fix_final=False)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', val=9.80665)

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        p['phase0.t_initial'] = 0.0
        p['phase0.t_duration'] = 2.0

        p['phase0.states:x'] = phase.interp('x', [0, 10])
        p['phase0.states:y'] = phase.interp('y', [10, 5])
        p['phase0.states:v'] = phase.interp('v', [0, 9.9])
        p['phase0.controls:theta'] = phase.interp('theta', [5, 100])
        p['phase0.parameters:g'] = 9.80665

        p.run_driver()

        exp_out = phase.simulate()

        t_initial = p.get_val('phase0.timeseries.time')[0]
        tf = p.get_val('phase0.timeseries.time')[-1]

        x0 = p.get_val('phase0.timeseries.states:x')[0]
        xf = p.get_val('phase0.timeseries.states:x')[-1]

        y0 = p.get_val('phase0.timeseries.states:y')[0]
        yf = p.get_val('phase0.timeseries.states:y')[-1]

        v0 = p.get_val('phase0.timeseries.states:v')[0]
        vf = p.get_val('phase0.timeseries.states:v')[-1]

        g = p.get_val('phase0.timeseries.parameters:g')[0]

        thetaf = exp_out.get_val('phase0.timeseries.controls:theta')[-1]

        assert_almost_equal(t_initial, 0.0)
        assert_almost_equal(x0, 0.0)
        assert_almost_equal(y0, 10.0)
        assert_almost_equal(v0, 0.0)

        assert_almost_equal(tf, 1.8016, decimal=4)
        assert_almost_equal(xf, 10.0, decimal=3)
        assert_almost_equal(yf, 5.0, decimal=3)
        assert_almost_equal(vf, 9.902, decimal=3)
        assert_almost_equal(g, 9.80665, decimal=3)

        assert_almost_equal(thetaf, 100.12, decimal=0)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_radau(self):
        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring()

        t = dm.Radau(num_segments=20,
                     order=[3, 5]*10,
                     compressed=True)

        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=t)

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False)
        phase.add_state('y', fix_initial=True, fix_final=False)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', val=9.80665)

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        p['phase0.t_initial'] = 0.0
        p['phase0.t_duration'] = 2.0

        p['phase0.states:x'] = phase.interp('x', [0, 10])
        p['phase0.states:y'] = phase.interp('y', [10, 5])
        p['phase0.states:v'] = phase.interp('v', [0, 9.9])
        p['phase0.controls:theta'] = phase.interp('theta', [5, 100])
        p['phase0.parameters:g'] = 9.80665

        p.run_driver()

        exp_out = phase.simulate()

        t_initial = p.get_val('phase0.timeseries.time')[0]
        tf = p.get_val('phase0.timeseries.time')[-1]

        x0 = p.get_val('phase0.timeseries.states:x')[0]
        xf = p.get_val('phase0.timeseries.states:x')[-1]

        y0 = p.get_val('phase0.timeseries.states:y')[0]
        yf = p.get_val('phase0.timeseries.states:y')[-1]

        v0 = p.get_val('phase0.timeseries.states:v')[0]
        vf = p.get_val('phase0.timeseries.states:v')[-1]

        g = p.get_val('phase0.timeseries.parameters:g')[0]

        thetaf = exp_out.get_val('phase0.timeseries.controls:theta')[-1]

        assert_almost_equal(t_initial, 0.0)
        assert_almost_equal(x0, 0.0)
        assert_almost_equal(y0, 10.0)
        assert_almost_equal(v0, 0.0)

        assert_almost_equal(tf, 1.8016, decimal=3)
        assert_almost_equal(xf, 10.0, decimal=3)
        assert_almost_equal(yf, 5.0, decimal=3)
        assert_almost_equal(vf, 9.902, decimal=3)
        assert_almost_equal(g, 9.80665, decimal=3)

        assert_almost_equal(thetaf, 100.12, decimal=0)



import unittest

import numpy as np
import openmdao.api as om
import dymos as dm

from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE
from dymos.transcriptions.explicit_shooting.rk_integration_comp import RKIntegrationComp

from openmdao.utils.testing_utils import require_pyoptsparse


class SimpleODE(om.ExplicitComponent):
    """
    A simple ODE from https://math.okstate.edu/people/yqwang/teaching/math4513_fall11/Notes/rungekutta.pdf
    """
    def initialize(self):
        self.options.declare('num_nodes', types=(int,))

    def setup(self):
        nn = self.options['num_nodes']
        self.add_input('x', shape=(nn,), units='s**2')
        self.add_input('t', shape=(nn,), units='s')
        self.add_input('p', shape=(nn,), units='s**2')

        self.add_output('x_dot', shape=(nn,), units='s')

        ar = np.arange(nn, dtype=int)
        self.declare_partials(of='x_dot', wrt='x', rows=ar, cols=ar, val=1.0)
        self.declare_partials(of='x_dot', wrt='t', rows=ar, cols=ar)
        self.declare_partials(of='x_dot', wrt='p', rows=ar, cols=ar, val=1.0)

    def compute(self, inputs, outputs):
        x = inputs['x']
        t = inputs['t']
        p = inputs['p']
        outputs['x_dot'] = x - t**2 + p

    def compute_partials(self, inputs, partials):
        t = inputs['t']
        partials['x_dot', 't'] = -2*t


class TestRKIntegrationComp(unittest.TestCase):

    def test_eval_f_scalar(self):
        gd = dm.transcriptions.grid_data.GridData(num_segments=10, transcription='gauss-lobatto',
                                                  transcription_order=3)

        time_options = dm.phase.options.TimeOptionsDictionary()

        time_options['targets'] = 't'
        time_options['units'] = 's'

        state_options = {'x': dm.phase.options.StateOptionsDictionary()}

        state_options['x']['shape'] = (1,)
        state_options['x']['units'] = 's**2'
        state_options['x']['rate_source'] = 'x_dot'
        state_options['x']['targets'] = ['x']

        param_options = {'p': dm.phase.options.ParameterOptionsDictionary()}

        param_options['p']['shape'] = (1,)
        param_options['p']['units'] = 's**2'
        param_options['p']['targets'] = ['p']

        control_options = {}
        polynomial_control_options = {}

        prob = om.Problem()

        prob.model.add_subsystem('fixed_step_integrator',
                                 RKIntegrationComp(SimpleODE, time_options, state_options,
                                                   param_options, control_options,
                                                   polynomial_control_options,
                                                   grid_data=gd,
                                                   num_steps_per_segment=100))
        prob.setup(mode='fwd', force_alloc_complex=True)

        prob.set_val('fixed_step_integrator.states:x', 0.5)
        prob.set_val('fixed_step_integrator.t_initial', 0.0)
        prob.set_val('fixed_step_integrator.t_duration', 2.0)
        prob.set_val('fixed_step_integrator.parameters:p', 1.0)

        x = np.array([[0.5]])
        t = np.array([[0.0]])
        theta = np.array([[0.0, 2.0, 1.0]]).T
        f = np.zeros_like(x)

        intg = prob.model.fixed_step_integrator
        intg._eval_subprob.model.ode_eval.set_segment_index(9)
        intg.eval_f(x, t, theta, f)

        assert_near_equal(f, x - t**2 + theta[2, 0])

    def test_eval_f_derivs_scalar(self):
        gd = dm.transcriptions.grid_data.GridData(num_segments=10, transcription='gauss-lobatto',
                                                  transcription_order=3)

        time_options = dm.phase.options.TimeOptionsDictionary()

        time_options['targets'] = 't'
        time_options['units'] = 's'

        state_options = {'x': dm.phase.options.StateOptionsDictionary()}

        state_options['x']['shape'] = (1,)
        state_options['x']['units'] = 's**2'
        state_options['x']['rate_source'] = 'x_dot'
        state_options['x']['targets'] = ['x']

        param_options = {'p': dm.phase.options.ParameterOptionsDictionary()}

        param_options['p']['shape'] = (1,)
        param_options['p']['units'] = 's**2'
        param_options['p']['targets'] = ['p']

        control_options = {}
        polynomial_control_options = {}

        prob = om.Problem()

        prob.model.add_subsystem('fixed_step_integrator',
                                 RKIntegrationComp(SimpleODE, time_options, state_options,
                                                   param_options, control_options,
                                                   polynomial_control_options,
                                                   grid_data=gd,
                                                   num_steps_per_segment=100))
        prob.setup(mode='fwd', force_alloc_complex=True)

        prob.set_val('fixed_step_integrator.states:x', 0.5)
        prob.set_val('fixed_step_integrator.t_initial', 0.0)
        prob.set_val('fixed_step_integrator.t_duration', 2.0)
        prob.set_val('fixed_step_integrator.parameters:p', 1.0)

        prob.final_setup()

        prob.run_model()

        x = np.array([[0.5]])
        t = np.array([[0.0]])
        theta = np.array([[0, 2.0, 1.0]]).T

        f_nom = np.zeros_like(x)

        f_x_fd = np.zeros_like(x)
        f_t_fd = np.zeros_like(x)
        f_theta_fd = np.zeros_like(x)

        intg = prob.model.fixed_step_integrator
        f_x = intg._f_x
        f_t = intg._f_t
        f_theta = intg._f_θ

        intg._eval_subprob.model.ode_eval.set_segment_index(9)

        intg.eval_f(x, t, theta, f_nom)
        intg.eval_f_derivs(x, t, theta, f_x, f_t, f_theta)

        step = 1.0E-8

        intg.eval_f(x + step, t, theta, f_x_fd)
        f_x_fd = (f_x_fd - f_nom) / step
        assert_near_equal(f_x, f_x_fd, tolerance=1.0E-6)

        intg.eval_f(x, t + step, theta, f_t_fd)
        f_t_fd = (f_t_fd - f_nom) / step
        assert_near_equal(f_t, f_t_fd, tolerance=1.0E-6)

        theta[2] = theta[2] + step
        prob.model.fixed_step_integrator.eval_f(x, t, theta, f_theta_fd)
        f_theta_fd = (f_theta_fd - f_nom) / step
        assert_near_equal(f_theta.real[0, 2], f_theta_fd[0, 0], tolerance=1.0E-6)

    def test_fwd_parameters(self):
        time_options = dm.phase.options.TimeOptionsDictionary()

        time_options['targets'] = 't'
        time_options['units'] = 's'

        state_options = {'x': dm.phase.options.StateOptionsDictionary()}

        state_options['x']['shape'] = (1,)
        state_options['x']['units'] = 's**2'
        state_options['x']['rate_source'] = 'x_dot'
        state_options['x']['targets'] = ['x']

        param_options = {'p': dm.phase.options.ParameterOptionsDictionary()}

        param_options['p']['shape'] = (1,)
        param_options['p']['units'] = 's**2'
        param_options['p']['targets'] = ['p']

        control_options = {}
        polynomial_control_options = {}

        p = om.Problem()

        gd = dm.transcriptions.grid_data.GridData(num_segments=10,
                                                  transcription='gauss-lobatto',
                                                  transcription_order=3,
                                                  compressed=True)

        p.model.add_subsystem('fixed_step_integrator', RKIntegrationComp(SimpleODE, time_options, state_options,
                                                                         param_options, control_options,
                                                                         polynomial_control_options,
                                                                         grid_data=gd,
                                                                         num_steps_per_segment=10,
                                                                         ode_init_kwargs=None))
        p.setup(mode='fwd', force_alloc_complex=True)

        p.set_val('fixed_step_integrator.states:x', 0.5)
        p.set_val('fixed_step_integrator.t_initial', 0.0)
        p.set_val('fixed_step_integrator.t_duration', 2.0)
        p.set_val('fixed_step_integrator.parameters:p', 1.0)

        p.run_model()

        t = p.get_val('fixed_step_integrator.time')
        x = p.get_val('fixed_step_integrator.states_out:x')

        assert_near_equal(t[-1, ...], 2)
        assert_near_equal(x[-1, ...], 5.305471950534675, tolerance=1.0E-7)

        with np.printoptions(linewidth=1024):
            cpd = p.check_partials(method='cs', compact_print=True)
            assert_check_partials(cpd)

    def test_fwd_parameters_controls(self):
        gd = dm.transcriptions.grid_data.GridData(num_segments=5, transcription='gauss-lobatto',
                                                  transcription_order=5, compressed=True)

        time_options = dm.phase.options.TimeOptionsDictionary()

        time_options['units'] = 's'

        state_options = {'x': dm.phase.options.StateOptionsDictionary(),
                         'y': dm.phase.options.StateOptionsDictionary(),
                         'v': dm.phase.options.StateOptionsDictionary()}

        state_options['x']['shape'] = (1,)
        state_options['x']['units'] = 'm'
        state_options['x']['rate_source'] = 'xdot'
        state_options['x']['targets'] = []

        state_options['y']['shape'] = (1,)
        state_options['y']['units'] = 'm'
        state_options['y']['rate_source'] = 'ydot'
        state_options['y']['targets'] = []

        state_options['v']['shape'] = (1,)
        state_options['v']['units'] = 'm/s'
        state_options['v']['rate_source'] = 'vdot'
        state_options['v']['targets'] = ['v']

        param_options = {'g': dm.phase.options.ParameterOptionsDictionary()}

        param_options['g']['shape'] = (1,)
        param_options['g']['units'] = 'm/s**2'
        param_options['g']['targets'] = ['g']

        control_options = {'theta': dm.phase.options.ControlOptionsDictionary()}

        control_options['theta']['shape'] = (1,)
        control_options['theta']['units'] = 'rad'
        control_options['theta']['targets'] = ['theta']

        polynomial_control_options = {}

        p = om.Problem()

        p.model.add_subsystem('fixed_step_integrator',
                              RKIntegrationComp(ode_class=BrachistochroneODE,
                                                time_options=time_options,
                                                state_options=state_options,
                                                parameter_options=param_options,
                                                control_options=control_options,
                                                polynomial_control_options=polynomial_control_options,
                                                num_steps_per_segment=10,
                                                grid_data=gd,
                                                ode_init_kwargs=None))

        p.setup(mode='fwd', force_alloc_complex=True)

        p.set_val('fixed_step_integrator.states:x', 0.0)
        p.set_val('fixed_step_integrator.states:y', 10.0)
        p.set_val('fixed_step_integrator.states:v', 0.0)
        p.set_val('fixed_step_integrator.t_initial', 0.0)
        p.set_val('fixed_step_integrator.t_duration', 1.8016)
        p.set_val('fixed_step_integrator.parameters:g', 9.80665)

        p.set_val('fixed_step_integrator.controls:theta', np.linspace(0.01, 100.0, 21), units='deg')

        p.run_model()

        x = p.get_val('fixed_step_integrator.states_out:x')
        y = p.get_val('fixed_step_integrator.states_out:y')
        v = p.get_val('fixed_step_integrator.states_out:v')
        theta = p.get_val('fixed_step_integrator.control_values:theta')

        # These tolerances are loose since theta is not properly spaced along the lgl nodes.
        assert_near_equal(x[-1, ...], 10.0, tolerance=0.1)
        assert_near_equal(y[-1, ...], 5.0, tolerance=0.1)
        assert_near_equal(v[-1, ...], 9.9, tolerance=0.1)

        with np.printoptions(linewidth=1024):
            cpd = p.check_partials(compact_print=True, method='cs', show_only_incorrect=True)
            assert_check_partials(cpd)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import io
import os

from packaging.version import Version

import numpy as np

from scipy.interpolate import interp1d

import openmdao.api as om
import openmdao.utils.assert_utils as _om_assert_utils
from openmdao import __version__ as openmdao_version


def assert_check_partials(data, atol=1.0E-6, rtol=1.0E-6):
    """
    Wrapper around OpenMDAO's assert_check_partials with a dymos-specific message.

    Calls OpenMDAO's assert_check_partials but verifies that the dictionary of assertion data is
    not empty due to dymos.options['include_check_partials'] being False.

    Parameters
    ----------
    data : dict of dicts of dicts
            First key:
                is the component name;
            Second key:
                is the (output, input) tuple of strings;
            Third key:
                is one of ['rel error', 'abs error', 'magnitude', 'J_fd', 'J_fwd', 'J_rev'];

            For 'rel error', 'abs error', 'magnitude' the value is: A tuple containing norms for
                forward - fd, adjoint - fd, forward - adjoint.
            For 'J_fd', 'J_fwd', 'J_rev' the value is: A numpy array representing the computed
                Jacobian for the three different methods of computation.
    atol : float
        Absolute error. Default is 1e-6.
    rtol : float
        Relative error. Default is 1e-6.
    """
    assert len(data) >= 1, "No check partials data found.  Is " \
                           "dymos.options['include_check_partials'] set to True?"
    _om_assert_utils.assert_check_partials(data, atol, rtol)


def assert_cases_equal(case1, case2, tol=1.0E-12, require_same_vars=True):
    """
    Raise AssertionError if the data in two OpenMDAO Cases is different.

    Parameters
    ----------
    case1 : om.Case
        The first OpenMDAO Case for comparison.
    case2 : om.Case
        The second OpenMDAO Case for comparison.
    tol : float
        The absolute value of the allowable difference in values between two variables.
    require_same_vars : bool
        If True, require that the two files contain the same set of variables.

    Raises
    ------
    AssertionError
        Raised in the following cases:  If require_same_vars is True, then AssertionError is raised
        if the two cases contain different variables.  Otherwise, this error is raised if case1
        and case2 contain the same variable but the variable has a different size/shape in the two
        cases, or if the variables have the same shape but different values (as given by tol).
    """
    _case1 = case1.model if isinstance(case1, om.Problem) else case1
    _case2 = case2.model if isinstance(case2, om.Problem) else case2

    case1_vars = {t[1]['prom_name']: t[1] for t in
                  _case1.list_inputs(val=True, units=True, prom_name=True, out_stream=None)}
    case1_vars.update({t[1]['prom_name']: t[1] for t in
                       _case1.list_outputs(val=True, units=True, prom_name=True, out_stream=None)})

    case2_vars = {t[1]['prom_name']: t[1] for t in
                  _case2.list_inputs(val=True, units=True, prom_name=True, out_stream=None)}
    case2_vars.update({t[1]['prom_name']: t[1] for t in
                       _case2.list_outputs(val=True, units=True, prom_name=True, out_stream=None)})

    # Warn if a and b don't contain the same sets of variables
    diff_err_msg = ''
    if require_same_vars:
        case1_minus_case2 = set(case1_vars.keys()) - set(case2_vars.keys())
        case2_minus_case1 = set(case2_vars.keys()) - set(case1_vars.keys())
        if case1_minus_case2 or case2_minus_case1:
            diff_err_msg = '\nrequire_same_vars=True but cases contain different variables.'
        if case1_minus_case2:
            diff_err_msg += f'\nVariables in case1 but not in case2: {sorted(case1_minus_case2)}'
        if case2_minus_case1:
            diff_err_msg += f'\nVariables in case2 but not in case1: {sorted(case2_minus_case1)}'

    shape_errors = set()
    val_errors = {}
    shape_err_msg = '\nThe following variables have different shapes/sizes:'
    val_err_msg = io.StringIO()

    for var in sorted(set(case1_vars.keys()).intersection(case2_vars.keys())):
        a = case1_vars[var]['val']
        b = case2_vars[var]['val']
        if a.shape != b.shape:
            shape_errors.add(var)
            shape_err_msg += f'\n{var} has shape {a.shape} in case1 but shape {b.shape} in case2'
            continue
        err = np.abs(a - b)
        max_err = np.max(err)
        mean_err = np.mean(err)
        if np.any(max_err > tol):
            val_errors[var] = (max_err, mean_err)

    err_msg = ''
    if diff_err_msg:
        err_msg += diff_err_msg
    if shape_errors:
        err_msg += shape_err_msg
    if val_errors:
        val_err_msg.write('\nThe following variables contain different values:\n')
        max_var_len = max(3, max([len(s) for s in val_errors.keys()]))
        val_err_msg.write(
            f"{'var'.rjust(max_var_len)} {'max error'.rjust(16)} {'mean error'.rjust(16)}\n")
        val_err_msg.write(max_var_len * '-' + ' ' + 16 * '-' + ' ' + 16 * '-' + '\n')
        for varname, (max_err, mean_err) in val_errors.items():
            val_err_msg.write(f"{varname.rjust(max_var_len)} {max_err:16.9e} {mean_err:16.9e}\n")
        err_msg += val_err_msg.getvalue()

    if err_msg:
        raise AssertionError(err_msg)


def _write_out_timeseries_values_out_of_tolerance(isclose, rel_tolerance, abs_tolerance,
                                                  t_check, x_check, x_ref):
    """
    Helper function used to write out a table of values indicating which timeseries values
    were out of tolerance.

    Parameters
    ----------
    isclose : array of bool
        Boolean array indicating where data value is in tolerance. Has same shape as the
        time series array
    rel_tolerance : float
        Allowed relative tolerance error
    abs_tolerance : float
        Allowed absolute tolerance error
    t_check : np.array
        Array of time values for the timeseries
    x_check : np.array
        Array of data values for the timeseries to be check/compared to the reference value, x_ref
    x_ref : np.array
        Array of data values for the timeseries to be used as the reference
    """
    err_msg = f"The following timeseries data are out of tolerance due to absolute (" \
              f"{abs_tolerance}) or relative ({rel_tolerance}) tolerance violations\n"
    header = f"{'time_index':10s} | " + \
             f"{'data_indices':12s} | " + \
             f"{'time':13s} | " + \
             f"{'ref_data':13s} | " + \
             f"{'checked_data':13s} | " + \
             f"{'abs_error':13s} | " + \
             f"{'rel_error':13} | " + \
             f" ABS or REL error "
    err_msg += f"{header}\n"
    err_msg += len(header) * '-' + '\n'

    rel_error_max = 0.0
    err_line_max = 0
    for idx, item_close in np.ndenumerate(isclose):
        if not item_close:
            error_string = ''
            abs_error = abs(x_check[idx] - x_ref[idx])
            if x_ref[idx] != 0.0:
                rel_error = abs(x_check[idx] - x_ref[idx]) / abs(x_ref[idx])
            else:
                rel_error = float('nan')
            if abs_tolerance is not None:
                if abs_error > abs_tolerance:
                    error_string += ' >ABS_TOL'

            if rel_tolerance is not None:
                if rel_error > rel_tolerance:
                    error_string += ' >REL_TOL'

            err_line = f"{idx[0]:10,d} | {str(idx[1:]):>12s} | {t_check[idx[0]]:13.6e} |" \
                       f"{x_ref[idx]:13.6e} | {x_check[idx]:13.6e} | {abs_error:13.6e} | " \
                       f"{rel_error:13.6e} | {error_string}\n"
            err_msg += err_line

            if rel_error > rel_error_max:
                rel_error_max = rel_error
                err_line_max = err_line

    # show the item with the max rel error
    max_rel_error_header_txt = 'Time series data value with the largest relative error'
    max_rel_error_msg = f"\n{len(max_rel_error_header_txt) * '#'}\n{max_rel_error_header_txt}\n" \
                        f"{len(max_rel_error_header_txt) * '#'}\n"

    max_rel_error_msg += f"{header}\n"
    max_rel_error_msg += len(header) * '-' + '\n'
    max_rel_error_msg += err_line_max

    err_msg += max_rel_error_msg

    return err_msg


def assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check, abs_tolerance=None,
                                 rel_tolerance=None):
    """
    Assert that two timeseries of data are approximately equal.

    The first timeseries, defined by t_ref, x_ref, serves as the reference.

    The second timeseries, defined by t_check, x_check is what is checked for near equality.

    The check is done by fitting a 1D interpolant to the reference, and then comparing
    the values of the interpolant at the times in t_check. The check for errors within
    tolerance are done on a point-by-point basis. If any point is out of tolerance, throw
    an AssertionError.

    Only the times where the two timeseries overlap are used for the check.

    When both abs_tolerance and rel_tolerance are given, only one is actually used for any given
    data point. When the absolute values of the data values are small, the abs_tolerance is used,
    otherwise the rel_tolerance is used. The transition point is given by

        abs_tolerance / rel_tolerance

    Parameters
    ----------
    t_ref : np.array
        Time values for the reference timeseries.
    x_ref : np.array
        Data values for the reference timeseries.
    t_check : np.array
        Time values for the timeseries that is compared to the reference.
    x_check : np.array
        Data values for the timeseries that is compared to the reference.
    abs_tolerance : float
        The absolute tolerance for any errors along at each point checked.
    rel_tolerance : float
        The relative tolerance for any errors along at each point checked.

    Raises
    ------
    AssertionError
        When one or more elements of the timeseries to be checked are not with in the desired
        tolerance of the interpolated reference timeseries, an AssertionError is raised.
    """
    # get shapes for the time series values
    shape_ref = x_ref.shape[1:]
    shape_check = x_check.shape[1:]

    if abs_tolerance is None and rel_tolerance is None:
        raise ValueError('abs_tolerance and rel_tolerance cannot be both None')

    if shape_ref != shape_check:
        raise ValueError('The shape of the variable in the two timeseries is not equal '
                         f'x_ref is {shape_ref}  x_check is {shape_check}')

    # get the overlapping time period between t_ref and t_check
    t_begin = max(t_ref[0], t_check[0])
    t_end = min(t_ref[-1], t_check[-1])

    if t_begin > t_end:
        raise ValueError("There is no overlapping time between the two time series")

    # Flatten the timeseries data arrays
    num_elements = np.prod(shape_ref, dtype=int)
    time_series_len = x_ref.shape[0]
    x_ref_data_flattened = np.reshape(x_ref, newshape=(time_series_len, num_elements))
    t_ref_unique, idxs_unique_ref = np.unique(t_ref.ravel(), return_index=True)
    x_to_interp = x_ref_data_flattened[idxs_unique_ref, ...]
    t_check = t_check.ravel()

    interp = interp1d(x=t_ref_unique, y=x_to_interp, kind='slinear', axis=0)

    # only want t_check in the overlapping range of t_begin and t_end
    t_check_in_range_condition = np.logical_and(t_check >= t_begin, t_check <= t_end)
    t_check = np.compress(t_check_in_range_condition, t_check)
    x_check = np.compress(t_check_in_range_condition, x_check, axis=0)

    # get the interpolated values of the reference at the values of t_check
    # Reshape back to unflattened data values
    x_ref_interp = np.reshape(interp(t_check), newshape=(t_check.size,) + shape_ref)

    if abs_tolerance is None:  # so only have rel_tolerance
        isclose = np.isclose(x_check, x_ref_interp, rtol=rel_tolerance, atol=0.0)
        all_close = np.all(isclose)
        if not all_close:
            err_msg = _write_out_timeseries_values_out_of_tolerance(isclose,
                                                                    rel_tolerance,
                                                                    abs_tolerance,
                                                                    t_check,
                                                                    x_check,
                                                                    x_ref_interp,
                                                                    )
            raise AssertionError(err_msg)
    elif rel_tolerance is None:  # so only have abs_tolerance
        isclose = np.isclose(x_check, x_ref_interp, rtol=0.0, atol=abs_tolerance)
        all_close = np.all(isclose)
        if not all_close:
            err_msg = _write_out_timeseries_values_out_of_tolerance(isclose,
                                                                    rel_tolerance,
                                                                    abs_tolerance,
                                                                    t_check,
                                                                    x_check,
                                                                    x_ref_interp,
                                                                    )
            raise AssertionError(err_msg)
    else:  # need to use a hybrid of abs and rel tolerances
        err_msg = ''

        # At what value of absolute value of the data does the tolerance check switch between
        #    using the absolute vs relative tolerance
        transition_tolerance = abs_tolerance / rel_tolerance

        # for values > transition_tolerance, use rel_tolerance
        transition_condition = abs(x_ref_interp) >= transition_tolerance
        above_transition_x_ref_interp = np.full(x_ref_interp.shape, np.nan)
        np.copyto(above_transition_x_ref_interp, x_ref_interp, where=transition_condition)
        above_transition_x_check = np.full(x_ref_interp.shape, np.nan)
        np.copyto(above_transition_x_check, x_check, where=transition_condition)
        isclose_using_rel_tolerance = np.isclose(above_transition_x_check,
                                                 above_transition_x_ref_interp,
                                                 rtol=rel_tolerance, atol=0.0, equal_nan=True)

        # for values < transition_tolerance, use abs_tolerance
        transition_condition = abs(x_ref_interp) < transition_tolerance
        below_transition_x_ref_interp = np.full(x_ref_interp.shape, np.nan)
        np.copyto(below_transition_x_ref_interp, x_ref_interp, where=transition_condition)
        below_transition_x_check = np.full(x_ref_interp.shape, np.nan)
        np.copyto(below_transition_x_check, x_check, where=transition_condition)
        isclose_using_abs_tolerance = np.isclose(below_transition_x_check,
                                                 below_transition_x_ref_interp, rtol=0.0,
                                                 atol=abs_tolerance, equal_nan=True)

        # combine the two
        isclose_using_both_tolerance = isclose_using_rel_tolerance & isclose_using_abs_tolerance
        all_close = np.all(isclose_using_both_tolerance)
        if not all_close:
            err_msg += _write_out_timeseries_values_out_of_tolerance(isclose_using_both_tolerance,
                                                                     rel_tolerance,
                                                                     abs_tolerance,
                                                                     t_check,
                                                                     x_check,
                                                                     x_ref_interp,
                                                                     )
        if err_msg:
            raise AssertionError(err_msg)


def _get_reports_dir(prob):
    # need this to work with older OM versions with old reports system API
    # reports API changed between 3.18 and 3.19, so handle it here in order to be able to
    #  test against older versions of openmdao
    if Version(openmdao_version) > Version("3.18"):
        return prob.get_reports_dir()

    from openmdao.utils.reports_system import get_reports_dir
    return get_reports_dir(prob)


# This duplicates OpenMDAO code and is needed for older versions of OpenMDAO (<= 3.19).
# Once support is dropped for < 3.19 we can get rid of this and use the version from OpenMDAO.
class set_env_vars(object):
    """
    Decorate a function to temporarily set some environment variables.

    Parameters
    ----------
    **envs : dict
        Keyword args corresponding to environment variables to set.

    Attributes
    ----------
    envs : dict
        Saved mapping of environment var name to value.
    """

    def __init__(self, **envs):
        """
        Initialize attributes.
        """
        self.envs = envs

    def __call__(self, fnc):
        """
        Apply the decorator.

        Parameters
        ----------
        fnc : function
            The function being wrapped.
        """
        def wrap(*args, **kwargs):
            saved = {}
            try:
                for k, v in self.envs.items():
                    saved[k] = os.environ.get(k)
                    os.environ[k] = v  # will raise exception if v is not a string

                return fnc(*args, **kwargs)
            finally:
                # put environment back as it was
                for k, v in saved.items():
                    if v is None:
                        del os.environ[k]
                    else:
                        os.environ[k] = v

        return wrap



import os
import unittest

import numpy as np
import matplotlib.pyplot as plt

from openmdao.api import Problem, Group, pyOptSparseDriver
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.general_utils import printoptions
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos import Trajectory, GaussLobatto, Phase, Radau
from dymos.examples.robot_arm.robot_arm_ode import RobotArmODE

plt.switch_backend('Agg')
show_plots = True


@use_tempdirs
@require_pyoptsparse(optimizer='SLSQP')
class TestRobotArm(unittest.TestCase):

    def tearDown(self):
        for filename in ['total_coloring.pkl', 'SLSQP.out', 'SNOPT_print.out']:
            if os.path.exists(filename):
                os.remove(filename)

    @require_pyoptsparse(optimizer='SLSQP')
    def make_problem(self, transcription=Radau, optimizer='SLSQP', numseg=30):
        p = Problem(model=Group())
        p.driver = pyOptSparseDriver()
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = optimizer
        if optimizer == 'SNOPT':
            p.driver.opt_settings['iSumm'] = 6
            p.driver.opt_settings['Verify level'] = 3
        elif optimizer == 'IPOPT':
            p.driver.opt_settings['nlp_scaling_method'] = 'gradient-based'
            p.driver.opt_settings['max_iter'] = 500
            p.driver.opt_settings['print_level'] = 0
        traj = p.model.add_subsystem('traj', Trajectory())

        phase = traj.add_phase('phase', Phase(ode_class=RobotArmODE,
                                              transcription=transcription(num_segments=numseg, order=3)))
        phase.set_time_options(fix_initial=True, fix_duration=False)

        phase.add_state('x0', fix_initial=True, fix_final=True, rate_source='x0_dot', units='m')
        phase.add_state('x1', fix_initial=True, fix_final=True, rate_source='x1_dot', units='rad')
        phase.add_state('x2', fix_initial=True, fix_final=True, rate_source='x2_dot', units='rad')
        phase.add_state('x3', fix_initial=True, fix_final=True, rate_source='x3_dot', units='m/s')
        phase.add_state('x4', fix_initial=True, fix_final=True, rate_source='x4_dot', units='rad/s')
        phase.add_state('x5', fix_initial=True, fix_final=True, rate_source='x5_dot', units='rad/s')

        phase.add_control('u0', opt=True, lower=-1, upper=1, scaler=0.1, units='m**2/s**2',
                          continuity=False, rate_continuity=False)
        phase.add_control('u1', opt=True, lower=-1, upper=1, scaler=0.1, units='m**3*rad/s**2',
                          continuity=False, rate_continuity=False)
        phase.add_control('u2', opt=True, lower=-1, upper=1, scaler=0.1, units='m**3*rad/s**2',
                          continuity=False, rate_continuity=False)

        phase.add_path_constraint('u0', lower=-1, upper=1, scaler=0.1)
        phase.add_path_constraint('u1', lower=-1, upper=1, scaler=0.1)
        phase.add_path_constraint('u2', lower=-1, upper=1, scaler=0.1)

        phase.add_objective('time', ref=0.1)

        phase.set_refine_options(refine=True, tol=1e-5, smoothness_factor=1.2)

        p.setup(check=True, force_alloc_complex=False, mode='auto')

        p.set_val('traj.phase.t_initial', 0)
        p.set_val('traj.phase.t_duration', 10)
        p.set_val('traj.phase.states:x0', phase.interp('x0', [4.5, 4.5]))
        p.set_val('traj.phase.states:x1', phase.interp('x1', [0.0, 2 * np.pi / 3]))
        p.set_val('traj.phase.states:x2', phase.interp('x2', [np.pi / 4, np.pi / 4]))
        p.set_val('traj.phase.states:x3', phase.interp('x3', [0.0, 0.0]))
        p.set_val('traj.phase.states:x4', phase.interp('x4', [0.0, 0.0]))
        p.set_val('traj.phase.states:x5', phase.interp('x5', [0.0, 0.0]))

        return p

    def test_partials(self):
        p = self.make_problem(transcription=Radau, optimizer='SLSQP')
        p.run_model()
        with printoptions(linewidth=1024, edgeitems=100):
            cpd = p.check_partials(method='fd', compact_print=True, out_stream=None)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_robot_arm_radau(self):
        p = self.make_problem(transcription=Radau, optimizer='IPOPT', numseg=12)
        dm.run_problem(p)

        t = p.get_val('traj.phase.timeseries.time')
        rho = p.get_val('traj.phase.timeseries.states:x0')
        theta = p.get_val('traj.phase.timeseries.states:x1')
        phi = p.get_val('traj.phase.timeseries.states:x2')
        u0 = p.get_val('traj.phase.timeseries.controls:u0')
        u1 = p.get_val('traj.phase.timeseries.controls:u1')
        u2 = p.get_val('traj.phase.timeseries.controls:u2')

        exp_out = p.model.traj.simulate()
        t_exp = exp_out.get_val('traj.phase.timeseries.time')
        rho_exp = exp_out.get_val('traj.phase.timeseries.states:x0')
        theta_exp = exp_out.get_val('traj.phase.timeseries.states:x1')
        phi_exp = exp_out.get_val('traj.phase.timeseries.states:x2')
        u0_exp = exp_out.get_val('traj.phase.timeseries.controls:u0')
        u1_exp = exp_out.get_val('traj.phase.timeseries.controls:u1')
        u2_exp = exp_out.get_val('traj.phase.timeseries.controls:u2')

        if show_plots:
            fig, axs = plt.subplots(2, 3)
            axs[0, 0].plot(t, rho, marker='o', linestyle='')
            axs[0, 1].plot(t, theta, marker='o', linestyle='')
            axs[0, 2].plot(t, phi, marker='o', linestyle='')
            axs[1, 0].plot(t, u0, marker='o', linestyle='')
            axs[1, 1].plot(t, u1, marker='o', linestyle='')
            axs[1, 2].plot(t, u2, marker='o', linestyle='')

            axs[0, 0].plot(t_exp, rho_exp, marker='', linestyle='-')
            axs[0, 1].plot(t_exp, theta_exp, marker='', linestyle='-')
            axs[0, 2].plot(t_exp, phi_exp, marker='', linestyle='-')
            axs[1, 0].plot(t_exp, u0_exp, marker='', linestyle='-')
            axs[1, 1].plot(t_exp, u1_exp, marker='', linestyle='-')
            axs[1, 2].plot(t_exp, u2_exp, marker='', linestyle='-')
            plt.show()

        assert_near_equal(t[-1], 9.14138, tolerance=1e-3)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_robot_arm_gl(self):
        p = self.make_problem(transcription=GaussLobatto, optimizer='IPOPT', numseg=20)
        dm.run_problem(p)

        t = p.get_val('traj.phase.timeseries.time')
        rho = p.get_val('traj.phase.timeseries.states:x0')
        theta = p.get_val('traj.phase.timeseries.states:x1')
        phi = p.get_val('traj.phase.timeseries.states:x2')
        u0 = p.get_val('traj.phase.timeseries.controls:u0')
        u1 = p.get_val('traj.phase.timeseries.controls:u1')
        u2 = p.get_val('traj.phase.timeseries.controls:u2')

        if show_plots:
            fig, axs = plt.subplots(2, 3)
            axs[0, 0].plot(t, rho)
            axs[0, 1].plot(t, theta)
            axs[0, 2].plot(t, phi)
            axs[1, 0].plot(t, u0)
            axs[1, 1].plot(t, u1)
            axs[1, 2].plot(t, u2)
            plt.show()

        assert_near_equal(t[-1], 9.14138, tolerance=1e-3)



import os
import unittest
import pathlib

import numpy as np

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos.examples.finite_burn_orbit_raise.finite_burn_eom import FiniteBurnODE
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE
from dymos.examples.battery_multibranch.battery_multibranch_ode import BatteryODE
from dymos.utils.lgl import lgl
from dymos.utils.testing_utils import _get_reports_dir
from dymos.visualization.timeseries_plots import timeseries_plots


@use_tempdirs
@require_pyoptsparse(optimizer='SLSQP')
class TestTimeSeriesPlotsBasics(unittest.TestCase):

    def setUp(self):
        optimizer = 'SLSQP'
        num_segments = 8
        transcription_order = 3
        compressed = True

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = optimizer
        p.driver.declare_coloring()

        t = dm.GaussLobatto(num_segments=num_segments,
                            order=transcription_order,
                            compressed=compressed)
        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=t)
        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj0', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False)
        phase.add_state('y', fix_initial=True, fix_final=False)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', val=9.80665)

        phase.add_timeseries('timeseries2',
                             transcription=dm.Radau(num_segments=num_segments * 5,
                                                    order=transcription_order,
                                                    compressed=compressed),
                             subset='control_input')

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=['unconnected_inputs'])

        p['traj0.phase0.t_initial'] = 0.0
        p['traj0.phase0.t_duration'] = 2.0

        p['traj0.phase0.states:x'] = phase.interp('x', [0, 10])
        p['traj0.phase0.states:y'] = phase.interp('y', [10, 5])
        p['traj0.phase0.states:v'] = phase.interp('v', [0, 9.9])
        p['traj0.phase0.controls:theta'] = phase.interp('theta', [5, 100])
        p['traj0.phase0.parameters:g'] = 9.80665

        p.setup()

        self.p = p

    def test_brachistochrone_timeseries_plots(self):
        dm.run_problem(self.p, make_plots=False)

        timeseries_plots('dymos_solution.db', problem=self.p)
        plot_dir = pathlib.Path(_get_reports_dir(self.p)).joinpath('plots')

        self.assertTrue(plot_dir.joinpath('states_x.png').exists())
        self.assertTrue(plot_dir.joinpath('states_y.png').exists())
        self.assertTrue(plot_dir.joinpath('states_v.png').exists())
        self.assertTrue(plot_dir.joinpath('controls_theta.png').exists())
        self.assertTrue(plot_dir.joinpath('control_rates_theta_rate.png').exists())
        self.assertTrue(plot_dir.joinpath('control_rates_theta_rate2.png').exists())

    def test_brachistochrone_timeseries_plots_solution_only_set_solution_record_file(self):
        # records to the default file 'dymos_simulation.db'
        dm.run_problem(self.p, make_plots=False, solution_record_file='solution_record_file.db')

        timeseries_plots('solution_record_file.db', problem=self.p)
        plot_dir = pathlib.Path(_get_reports_dir(self.p)).joinpath('plots')

        self.assertTrue(plot_dir.joinpath('states_x.png').exists())
        self.assertTrue(plot_dir.joinpath('states_y.png').exists())
        self.assertTrue(plot_dir.joinpath('states_v.png').exists())
        self.assertTrue(plot_dir.joinpath('controls_theta.png').exists())
        self.assertTrue(plot_dir.joinpath('control_rates_theta_rate.png').exists())
        self.assertTrue(plot_dir.joinpath('control_rates_theta_rate2.png').exists())

    def test_brachistochrone_timeseries_plots_solution_and_simulation(self):
        dm.run_problem(self.p, simulate=True, make_plots=False,
                       simulation_record_file='simulation_record_file.db')

        timeseries_plots('dymos_solution.db', simulation_record_file='simulation_record_file.db', problem=self.p)

    def test_brachistochrone_timeseries_plots_set_plot_dir(self):
        dm.run_problem(self.p, make_plots=False)

        plot_dir = pathlib.Path(_get_reports_dir(self.p)).joinpath("test_plot_dir").resolve()
        timeseries_plots('dymos_solution.db', plot_dir=plot_dir, problem=self.p)

        self.assertTrue(plot_dir.joinpath('states_x.png').exists())
        self.assertTrue(plot_dir.joinpath('states_y.png').exists())
        self.assertTrue(plot_dir.joinpath('states_v.png').exists())
        self.assertTrue(plot_dir.joinpath('controls_theta.png').exists())
        self.assertTrue(plot_dir.joinpath('control_rates_theta_rate.png').exists())
        self.assertTrue(plot_dir.joinpath('control_rates_theta_rate2.png').exists())


@use_tempdirs
class TestTimeSeriesPlotsMultiPhase(unittest.TestCase):

    @require_pyoptsparse(optimizer='IPOPT')
    def test_trajectory_linked_phases_make_plot(self):

        self.traj = dm.Trajectory()
        p = self.p = om.Problem(model=self.traj)

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'IPOPT'
        p.driver.declare_coloring()

        # First Phase (burn)
        burn1 = dm.Phase(ode_class=FiniteBurnODE, transcription=dm.GaussLobatto(num_segments=4,
                                                                                order=3))

        self.traj.add_phase('burn1', burn1)

        burn1.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='TU')
        burn1.add_state('r', fix_initial=True, fix_final=False,
                        rate_source='r_dot', targets=['r'], units='DU')
        burn1.add_state('theta', fix_initial=True, fix_final=False,
                        rate_source='theta_dot', targets=['theta'], units='rad')
        burn1.add_state('vr', fix_initial=True, fix_final=False,
                        rate_source='vr_dot', targets=['vr'], units='DU/TU')
        burn1.add_state('vt', fix_initial=True, fix_final=False,
                        rate_source='vt_dot', targets=['vt'], units='DU/TU')
        burn1.add_state('accel', fix_initial=True, fix_final=False,
                        rate_source='at_dot', targets=['accel'], units='DU/TU**2')
        burn1.add_state('deltav', fix_initial=True, fix_final=False,
                        rate_source='deltav_dot', units='DU/TU')
        burn1.add_control('u1',  targets=['u1'], rate_continuity=True, rate2_continuity=True,
                          units='deg', scaler=0.01, lower=-30, upper=30)
        burn1.add_parameter('c', opt=False, val=1.5, targets=['c'], units='DU/TU')

        # Second Phase (Coast)

        coast = dm.Phase(ode_class=FiniteBurnODE, transcription=dm.GaussLobatto(num_segments=10,
                                                                                order=3))

        self.traj.add_phase('coast', coast)

        coast.set_time_options(initial_bounds=(0.5, 20), duration_bounds=(.5, 10), duration_ref=10,
                               units='TU')
        coast.add_state('r', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='r_dot', targets=['r'], units='DU')
        coast.add_state('theta', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='theta_dot', targets=['theta'], units='rad')
        coast.add_state('vr', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='vr_dot', targets=['vr'], units='DU/TU')
        coast.add_state('vt', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='vt_dot', targets=['vt'], units='DU/TU')
        coast.add_state('accel', fix_initial=True, fix_final=True, ref=1.0E-12, defect_ref=1.0E-12,
                        rate_source='at_dot', targets=['accel'], units='DU/TU**2')
        coast.add_state('deltav', fix_initial=False, fix_final=False,
                        rate_source='deltav_dot', units='DU/TU')
        coast.add_control('u1', targets=['u1'], opt=False, val=0.0, units='deg')
        coast.add_parameter('c', opt=False, val=1.5, targets=['c'], units='DU/TU')

        # Third Phase (burn)

        burn2 = dm.Phase(ode_class=FiniteBurnODE, transcription=dm.GaussLobatto(num_segments=3,
                                                                                order=3))

        self.traj.add_phase('burn2', burn2)

        burn2.set_time_options(initial_bounds=(0.5, 20), duration_bounds=(.5, 10), initial_ref=10,
                               units='TU')
        burn2.add_state('r', fix_initial=False, fix_final=True,
                        rate_source='r_dot', targets=['r'], units='DU')
        burn2.add_state('theta', fix_initial=False, fix_final=False,
                        rate_source='theta_dot', targets=['theta'], units='rad')
        burn2.add_state('vr', fix_initial=False, fix_final=True,
                        rate_source='vr_dot', targets=['vr'], units='DU/TU')
        burn2.add_state('vt', fix_initial=False, fix_final=True,
                        rate_source='vt_dot', targets=['vt'], units='DU/TU')
        burn2.add_state('accel', fix_initial=False, fix_final=False, defect_ref=1.0E-6,
                        rate_source='at_dot', targets=['accel'], units='DU/TU**2')
        burn2.add_state('deltav', fix_initial=False, fix_final=False,
                        rate_source='deltav_dot', units='DU/TU')
        burn2.add_control('u1', targets=['u1'], rate_continuity=True, rate2_continuity=True,
                          units='deg', scaler=0.01, lower=-30, upper=30)
        burn2.add_parameter('c', opt=False, val=1.5, targets=['c'], units='DU/TU')

        burn2.add_objective('deltav', loc='final')

        # Link Phases
        self.traj.link_phases(phases=['burn1', 'coast', 'burn2'],
                              vars=['time', 'r', 'theta', 'vr', 'vt', 'deltav'])
        self.traj.link_phases(phases=['burn1', 'burn2'], vars=['accel'])

        # Finish Problem Setup
        p.model.linear_solver = om.DirectSolver()

        p.setup(check=True)

        # Set Initial Guesses

        p.set_val('burn1.t_initial', value=0.0)
        p.set_val('burn1.t_duration', value=2.25)

        p.set_val('burn1.states:r', value=burn1.interp('r', [1, 1.5]))
        p.set_val('burn1.states:theta', value=burn1.interp('theta', [0, 1.7]))
        p.set_val('burn1.states:vr', value=burn1.interp('vr', [0, 0]))
        p.set_val('burn1.states:vt', value=burn1.interp('vt', [1, 1]))
        p.set_val('burn1.states:accel', value=burn1.interp('accel', [0.1, 0]))
        p.set_val('burn1.states:deltav', value=burn1.interp('deltav', [0, 0.1]))
        p.set_val('burn1.controls:u1', value=burn1.interp('u1', [-3.5, 13.0]))
        p.set_val('burn1.parameters:c', value=1.5)

        p.set_val('coast.t_initial', value=2.25)
        p.set_val('coast.t_duration', value=3.0)

        p.set_val('coast.states:r', value=coast.interp('r', [1.3, 1.5]))
        p.set_val('coast.states:theta', value=coast.interp('theta', [2.1767, 1.7]))
        p.set_val('coast.states:vr', value=coast.interp('vr', [0.3285, 0]))
        p.set_val('coast.states:vt', value=coast.interp('vt', [0.97, 1]))
        p.set_val('coast.states:accel', value=coast.interp('accel', [0, 0]))
        p.set_val('coast.controls:u1', value=coast.interp('u1', [0, 0]))
        p.set_val('coast.parameters:c', value=1.5)

        p.set_val('burn2.t_initial', value=5.25)
        p.set_val('burn2.t_duration', value=1.75)

        p.set_val('burn2.states:r', value=burn2.interp('r', [1, 3]))
        p.set_val('burn2.states:theta', value=burn2.interp('theta', [0, 4.0]))
        p.set_val('burn2.states:vr', value=burn2.interp('vr', [0, 0]))
        p.set_val('burn2.states:vt', value=burn2.interp('vt', [1, np.sqrt(1 / 3)]))
        p.set_val('burn2.states:accel', value=burn2.interp('accel', [0.1, 0]))
        p.set_val('burn2.states:deltav', value=burn2.interp('deltav', [0.1, 0.2]))
        p.set_val('burn2.controls:u1', value=burn2.interp('u1', [1, 1]))
        p.set_val('burn2.parameters:c', value=1.5)

        dm.run_problem(p, simulate=True, make_plots=False,
                       simulation_record_file='simulation_record_file.db')

        timeseries_plots('dymos_solution.db', simulation_record_file='simulation_record_file.db',
                         problem=p)
        plot_dir = pathlib.Path(_get_reports_dir(p)).joinpath("plots")

        for varname in ['time_phase', 'states:r', 'state_rates:r', 'states:theta',
                        'state_rates:theta', 'states:vr', 'state_rates:vr', 'states:vt',
                        'state_rates:vt', 'states:accel',
                        'state_rates:accel', 'states:deltav', 'state_rates:deltav',
                        'controls:u1', 'control_rates:u1_rate', 'control_rates:u1_rate2',
                        'parameters:c']:
            self.assertTrue(plot_dir.joinpath(varname.replace(":", "_") + '.png').exists())

    def test_overlapping_phases_make_plot(self):

        prob = om.Problem()

        opt = prob.driver = om.ScipyOptimizeDriver()
        opt.declare_coloring()
        opt.options['optimizer'] = 'SLSQP'

        num_seg = 5
        seg_ends, _ = lgl(num_seg + 1)

        traj = prob.model.add_subsystem('traj', dm.Trajectory())

        # First phase: normal operation.
        transcription = dm.Radau(num_segments=num_seg, order=5, segment_ends=seg_ends,
                                 compressed=False)
        phase0 = dm.Phase(ode_class=BatteryODE, transcription=transcription)
        traj_p0 = traj.add_phase('phase0', phase0)

        traj_p0.set_time_options(fix_initial=True, fix_duration=True)
        traj_p0.add_state('state_of_charge', fix_initial=True, fix_final=False,
                          targets=['SOC'], rate_source='dXdt:SOC')

        # Second phase: normal operation.

        phase1 = dm.Phase(ode_class=BatteryODE, transcription=transcription)
        traj_p1 = traj.add_phase('phase1', phase1)

        traj_p1.set_time_options(fix_initial=False, fix_duration=True)
        traj_p1.add_state('state_of_charge', fix_initial=False, fix_final=False,
                          targets=['SOC'], rate_source='dXdt:SOC')
        traj_p1.add_objective('time', loc='final')

        # Second phase, but with battery failure.

        phase1_bfail = dm.Phase(ode_class=BatteryODE, ode_init_kwargs={'num_battery': 2},
                                transcription=transcription)
        traj_p1_bfail = traj.add_phase('phase1_bfail', phase1_bfail)

        traj_p1_bfail.set_time_options(fix_initial=False, fix_duration=True)
        traj_p1_bfail.add_state('state_of_charge', fix_initial=False, fix_final=False,
                                targets=['SOC'], rate_source='dXdt:SOC')

        # Second phase, but with motor failure.

        phase1_mfail = dm.Phase(ode_class=BatteryODE, ode_init_kwargs={'num_motor': 2},
                                transcription=transcription)
        traj_p1_mfail = traj.add_phase('phase1_mfail', phase1_mfail)

        traj_p1_mfail.set_time_options(fix_initial=False, fix_duration=True)
        traj_p1_mfail.add_state('state_of_charge', fix_initial=False, fix_final=False,
                                targets=['SOC'], rate_source='dXdt:SOC')

        traj.link_phases(phases=['phase0', 'phase1'], vars=['state_of_charge', 'time'])
        traj.link_phases(phases=['phase0', 'phase1_bfail'], vars=['state_of_charge', 'time'])
        traj.link_phases(phases=['phase0', 'phase1_mfail'], vars=['state_of_charge', 'time'])

        prob.model.options['assembled_jac_type'] = 'csc'
        prob.model.linear_solver = om.DirectSolver(assemble_jac=True)

        prob.setup()

        prob['traj.phase0.t_initial'] = 0
        prob['traj.phase0.t_duration'] = 1.0*3600

        prob['traj.phase1.t_initial'] = 1.0*3600
        prob['traj.phase1.t_duration'] = 1.0*3600

        prob['traj.phase1_bfail.t_initial'] = 1.0*3600
        prob['traj.phase1_bfail.t_duration'] = 1.0*3600

        prob['traj.phase1_mfail.t_initial'] = 1.0*3600
        prob['traj.phase1_mfail.t_duration'] = 1.0*3600

        prob.set_solver_print(level=0)
        dm.run_problem(prob, simulate=True, make_plots=False,
                       simulation_record_file='simulation_record_file.db')

        timeseries_plots('dymos_solution.db', simulation_record_file='simulation_record_file.db',
                         problem=prob)
        plot_dir = pathlib.Path(_get_reports_dir(prob)).joinpath("plots")

        self.assertTrue(plot_dir.joinpath('time_phase.png').exists())
        self.assertTrue(plot_dir.joinpath('states_state_of_charge.png').exists())
        self.assertTrue(plot_dir.joinpath('state_rates_state_of_charge.png').exists())

    @require_pyoptsparse(optimizer='IPOPT')
    def test_trajectory_linked_phases_make_plot_missing_data(self):
        """
        Test that plots are still generated even if the phases don't share the exact same
        variables in the timeseries.
        """

        self.traj = dm.Trajectory()
        p = self.p = om.Problem(model=self.traj)

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'IPOPT'
        p.driver.declare_coloring()

        # First Phase (burn)
        burn1 = dm.Phase(ode_class=FiniteBurnODE, transcription=dm.GaussLobatto(num_segments=4,
                                                                                order=3))

        self.traj.add_phase('burn1', burn1)

        burn1.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='TU')
        burn1.add_state('r', fix_initial=True, fix_final=False,
                        rate_source='r_dot', targets=['r'], units='DU')
        burn1.add_state('theta', fix_initial=True, fix_final=False,
                        rate_source='theta_dot', targets=['theta'], units='rad')
        burn1.add_state('vr', fix_initial=True, fix_final=False,
                        rate_source='vr_dot', targets=['vr'], units='DU/TU')
        burn1.add_state('vt', fix_initial=True, fix_final=False,
                        rate_source='vt_dot', targets=['vt'], units='DU/TU')
        burn1.add_state('accel', fix_initial=True, fix_final=False,
                        rate_source='at_dot', targets=['accel'], units='DU/TU**2')
        burn1.add_state('deltav', fix_initial=True, fix_final=False,
                        rate_source='deltav_dot', units='DU/TU')
        burn1.add_control('u1',  targets=['u1'], rate_continuity=True, rate2_continuity=True,
                          units='deg', scaler=0.01, lower=-30, upper=30)
        burn1.add_parameter('c', opt=False, val=1.5, targets=['c'], units='DU/TU')

        # Second Phase (Coast)

        coast = dm.Phase(ode_class=FiniteBurnODE, transcription=dm.GaussLobatto(num_segments=10,
                                                                                order=3))

        self.traj.add_phase('coast', coast)

        coast.set_time_options(initial_bounds=(0.5, 20), duration_bounds=(.5, 10), duration_ref=10,
                               units='TU')
        coast.add_state('r', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='r_dot', targets=['r'], units='DU')
        coast.add_state('theta', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='theta_dot', targets=['theta'], units='rad')
        coast.add_state('vr', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='vr_dot', targets=['vr'], units='DU/TU')
        coast.add_state('vt', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='vt_dot', targets=['vt'], units='DU/TU')
        coast.add_state('accel', fix_initial=True, fix_final=True, ref=1.0E-12, defect_ref=1.0E-12,
                        rate_source='at_dot', targets=['accel'], units='DU/TU**2')
        coast.add_state('deltav', fix_initial=False, fix_final=False,
                        rate_source='deltav_dot', units='DU/TU')
        coast.add_parameter('u1', targets=['u1'], opt=False, val=0.0, units='deg')
        coast.add_parameter('c', opt=False, val=1.5, targets=['c'], units='DU/TU')

        # Third Phase (burn)

        burn2 = dm.Phase(ode_class=FiniteBurnODE, transcription=dm.GaussLobatto(num_segments=3,
                                                                                order=3))

        self.traj.add_phase('burn2', burn2)

        burn2.set_time_options(initial_bounds=(0.5, 20), duration_bounds=(.5, 10), initial_ref=10,
                               units='TU')
        burn2.add_state('r', fix_initial=False, fix_final=True,
                        rate_source='r_dot', targets=['r'], units='DU')
        burn2.add_state('theta', fix_initial=False, fix_final=False,
                        rate_source='theta_dot', targets=['theta'], units='rad')
        burn2.add_state('vr', fix_initial=False, fix_final=True,
                        rate_source='vr_dot', targets=['vr'], units='DU/TU')
        burn2.add_state('vt', fix_initial=False, fix_final=True,
                        rate_source='vt_dot', targets=['vt'], units='DU/TU')
        burn2.add_state('accel', fix_initial=False, fix_final=False, defect_ref=1.0E-6,
                        rate_source='at_dot', targets=['accel'], units='DU/TU**2')
        burn2.add_state('deltav', fix_initial=False, fix_final=False,
                        rate_source='deltav_dot', units='DU/TU')
        burn2.add_control('u1', targets=['u1'], rate_continuity=True, rate2_continuity=True,
                          units='deg', scaler=0.01, lower=-30, upper=30)
        burn2.add_parameter('c', opt=False, val=1.5, targets=['c'], units='DU/TU')

        burn2.add_objective('deltav', loc='final')

        # Link Phases
        self.traj.link_phases(phases=['burn1', 'coast', 'burn2'],
                              vars=['time', 'r', 'theta', 'vr', 'vt', 'deltav'])
        self.traj.link_phases(phases=['burn1', 'burn2'], vars=['accel'])

        # Finish Problem Setup
        p.model.linear_solver = om.DirectSolver()

        p.setup(check=True)

        # Set Initial Guesses
        p.set_val('burn1.t_initial', value=0.0)
        p.set_val('burn1.t_duration', value=2.25)

        p.set_val('burn1.states:r', value=burn1.interp('r', [1, 1.5]))
        p.set_val('burn1.states:theta', value=burn1.interp('theta', [0, 1.7]))
        p.set_val('burn1.states:vr', value=burn1.interp('vr', [0, 0]))
        p.set_val('burn1.states:vt', value=burn1.interp('vt', [1, 1]))
        p.set_val('burn1.states:accel', value=burn1.interp('accel', [0.1, 0]))
        p.set_val('burn1.states:deltav', value=burn1.interp('deltav', [0, 0.1]))
        p.set_val('burn1.controls:u1', value=burn1.interp('u1', [-3.5, 13.0]))
        p.set_val('burn1.parameters:c', value=1.5)

        p.set_val('coast.t_initial', value=2.25)
        p.set_val('coast.t_duration', value=3.0)

        p.set_val('coast.states:r', value=coast.interp('r', [1.3, 1.5]))
        p.set_val('coast.states:theta', value=coast.interp('theta', [2.1767, 1.7]))
        p.set_val('coast.states:vr', value=coast.interp('vr', [0.3285, 0]))
        p.set_val('coast.states:vt', value=coast.interp('vt', [0.97, 1]))
        p.set_val('coast.states:accel', value=coast.interp('accel', [0, 0]))
        p.set_val('coast.parameters:c', value=1.5)

        p.set_val('burn2.t_initial', value=5.25)
        p.set_val('burn2.t_duration', value=1.75)

        p.set_val('burn2.states:r', value=burn2.interp('r', [1, 3]))
        p.set_val('burn2.states:theta', value=burn2.interp('theta', [0, 4.0]))
        p.set_val('burn2.states:vr', value=burn2.interp('vr', [0, 0]))
        p.set_val('burn2.states:vt', value=burn2.interp('vt', [1, np.sqrt(1 / 3)]))
        p.set_val('burn2.states:accel', value=burn2.interp('accel', [0.1, 0]))
        p.set_val('burn2.states:deltav', value=burn2.interp('deltav', [0.1, 0.2]))
        p.set_val('burn2.controls:u1', value=burn2.interp('u1', [1, 1]))
        p.set_val('burn2.parameters:c', value=1.5)

        dm.run_problem(p, simulate=True, make_plots=False,
                       simulation_record_file='simulation_record_file.db')

        timeseries_plots('dymos_solution.db', simulation_record_file='simulation_record_file.db',
                         problem=p)
        plot_dir = pathlib.Path(_get_reports_dir(p)).joinpath("plots")

        for varname in ['time_phase', 'states:r', 'state_rates:r', 'states:theta',
                        'state_rates:theta', 'states:vr', 'state_rates:vr', 'states:vt',
                        'state_rates:vt', 'states:accel',
                        'state_rates:accel', 'states:deltav', 'state_rates:deltav',
                        'controls:u1', 'control_rates:u1_rate', 'control_rates:u1_rate2',
                        'parameters:c', 'parameters:u1']:
            self.assertTrue(plot_dir.joinpath(varname.replace(":", "_") + '.png').exists())


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



from collections import OrderedDict
from collections.abc import Sequence
from copy import deepcopy
import itertools
import sys
import warnings
try:
    from itertools import izip
except ImportError:
    izip = zip

import numpy as np

import openmdao.api as om
from openmdao.utils.mpi import MPI

from ..utils.constants import INF_BOUND

from .options import LinkageOptionsDictionary
from .phase_linkage_comp import PhaseLinkageComp
from ..phase.analytic_phase import AnalyticPhase
from ..phase.options import TrajParameterOptionsDictionary
from ..transcriptions.common import ParameterComp
from ..utils.misc import get_rate_units, _unspecified
from ..utils.introspection import get_promoted_vars, get_source_metadata


class Trajectory(om.Group):
    """
    Class definition for a Trajectory group.

    A Trajectory object serves as a container for one or more Phases, as well as the linkage
    conditions between phases.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.

    Attributes
    ----------
    parameter_options : dict
        A dictionary of parameter names and their associated TrajectoryParameterOptionsDictionary
    phases : om.Group or om.ParallelGroup
        The Group which contains phases for this Trajectory.

    _linkages : OrderedDict
        A dictionary containing phase linkage information for the Trajectory.
    _phases : dict
        A dictionary of phase names as keys with the Phase objects being their associated values.
    """
    def __init__(self, **kwargs):
        super(Trajectory, self).__init__(**kwargs)

        self._linkages = {}
        self._phases = {}

        self.parameter_options = {}
        self.phases = om.ParallelGroup()

    def initialize(self):
        """
        Declare any options for Trajectory.
        """
        self.options.declare('sim_mode', types=bool, default=False,
                             desc='Used internally by Dymos when invoking simulate on a trajectory')

    def add_phase(self, name, phase, **kwargs):
        """
        Add a phase to the trajectory.

        Phases will be added to the Trajectory's `phases` subgroup.

        Parameters
        ----------
        name : str
            The name of the phase being added.
        phase : dymos Phase object
            The Phase object to be added.
        **kwargs : dict
            Additional arguments when adding the phase to the trajectory.

        Returns
        -------
        PhaseBase
            The Phase object added to the trajectory.
        """
        self._phases[name] = self.phases.add_subsystem(name, phase, **kwargs)
        return phase

    def set_parameter_options(self, name, units=_unspecified, val=_unspecified, desc=_unspecified, opt=False,
                              targets=_unspecified, lower=_unspecified, upper=_unspecified,
                              scaler=_unspecified, adder=_unspecified, ref0=_unspecified, ref=_unspecified,
                              shape=_unspecified, dynamic=_unspecified, static_target=_unspecified):
        """
        Set the options of an existing a parameter in the trajectory.

        Parameters
        ----------
        name : str
            Name of the parameter.
        units : str or None or _unspecified
            Units in which the parameter is defined.  If _unspecified, use the units declared
            for the parameter in the ODE.
        val : float or ndarray
            Default value of the parameter at all nodes.
        desc : str
            A description of the parameter.
        opt : bool
            If True the value(s) of this parameter will be design variables in
            the optimization problem. The default is False.
        targets : dict or None
            If None, then the parameter will be connected to the controllable parameter
            in the ODE of each phase.  For each phase where no such controllable parameter exists,
            a warning will be issued.  If targets is given as a dict, the dict should provide
            the relevant phase names as keys, each associated with the respective controllable
            parameter as a value.
        lower : float or ndarray
            The lower bound of the parameter value.
        upper : float or ndarray
            The upper bound of the parameter value.
        scaler : float or ndarray
            The scaler of the parameter value for the optimizer.
        adder : float or ndarray
            The adder of the parameter value for the optimizer.
        ref0 : float or ndarray
            The zero-reference value of the parameter for the optimizer.
        ref : float or ndarray
            The unit-reference value of the parameter for the optimizer.
        shape : Sequence of int
            The shape of the parameter.
        dynamic : bool
            True if the targets in the ODE may be dynamic (if the inputs are sized to the number
            of nodes) else False. This argument is deprecated in favor of static_target.
        static_target : bool or _unspecified
            True if the targets in the ODE are not shaped with num_nodes as the first dimension
            (meaning they cannot have a unique value at each node).  Otherwise False.
        """
        if name not in self.parameter_options:
            self.parameter_options[name] = TrajParameterOptionsDictionary()

        if units is not _unspecified:
            self.parameter_options[name]['units'] = units

        if opt is not _unspecified:
            self.parameter_options[name]['opt'] = opt

        if val is not _unspecified:
            self.parameter_options[name]['val'] = val

        if desc is not _unspecified:
            self.parameter_options[name]['desc'] = desc

        if lower is not _unspecified:
            self.parameter_options[name]['lower'] = lower

        if upper is not _unspecified:
            self.parameter_options[name]['upper'] = upper

        if scaler is not _unspecified:
            self.parameter_options[name]['scaler'] = scaler

        if adder is not _unspecified:
            self.parameter_options[name]['adder'] = adder

        if ref0 is not _unspecified:
            self.parameter_options[name]['ref0'] = ref0

        if ref is not _unspecified:
            self.parameter_options[name]['ref'] = ref

        if targets is not _unspecified:
            if isinstance(targets, str):
                self.parameter_options[name]['targets'] = (targets,)
            else:
                self.parameter_options[name]['targets'] = targets

        if shape is not _unspecified:
            self.parameter_options[name]['shape'] = shape

        if dynamic is not _unspecified:
            self.parameter_options[name]['static_target'] = not dynamic

        if static_target is not _unspecified:
            self.parameter_options[name]['static_target'] = static_target

        if dynamic is not _unspecified and static_target is not _unspecified:
            raise ValueError("Both the deprecated 'dynamic' option and option 'static_target' were "
                             f"specified for parameter '{name}'. "
                             f"Going forward, please use only option static_target.  Option "
                             f"'dynamic' will be removed in Dymos 2.0.0.")

    def add_parameter(self, name, units=_unspecified, val=_unspecified, desc=_unspecified, opt=False,
                      targets=_unspecified, lower=_unspecified, upper=_unspecified,
                      scaler=_unspecified, adder=_unspecified, ref0=_unspecified, ref=_unspecified,
                      shape=_unspecified, dynamic=_unspecified, static_target=_unspecified):
        """
        Add a parameter (static control) to the trajectory.

        Parameters
        ----------
        name : str
            Name of the parameter.
        units : str or None or _unspecified
            Units in which the parameter is defined.  If _unspecified, use the units declared
            for the parameter in the ODE.
        val : float or ndarray
            Default value of the parameter at all nodes.
        desc : str
            A description of the parameter.
        opt : bool
            If True the value(s) of this parameter will be design variables in
            the optimization problem. The default is False.
        targets : dict or None
            If None, then the parameter will be connected to the controllable parameter
            in the ODE of each phase.  For each phase where no such controllable parameter exists,
            a warning will be issued.  If targets is given as a dict, the dict should provide
            the relevant phase names as keys, each associated with the respective controllable
            parameter as a value.
        lower : float or ndarray
            The lower bound of the parameter value.
        upper : float or ndarray
            The upper bound of the parameter value.
        scaler : float or ndarray
            The scaler of the parameter value for the optimizer.
        adder : float or ndarray
            The adder of the parameter value for the optimizer.
        ref0 : float or ndarray
            The zero-reference value of the parameter for the optimizer.
        ref : float or ndarray
            The unit-reference value of the parameter for the optimizer.
        shape : Sequence of int
            The shape of the parameter.
        dynamic : bool
            True if the targets in the ODE may be dynamic (if the inputs are sized to the number
            of nodes) else False. This argument is deprecated in favor of static_target.
        static_target : bool or _unspecified
            True if the targets in the ODE are not shaped with num_nodes as the first dimension
            (meaning they cannot have a unique value at each node).  Otherwise False.
        """
        if name not in self.parameter_options:
            self.parameter_options[name] = TrajParameterOptionsDictionary()
            self.parameter_options[name]['name'] = name
        else:
            raise ValueError(f'Attempted to add parameter "{name}" to trajectory but trajectory already has a parameter'
                             'of that name.')

        self.set_parameter_options(name, units=units, val=val, desc=desc, opt=opt, targets=targets, lower=lower,
                                   upper=upper, scaler=scaler, adder=adder, ref0=ref0, ref=ref, shape=shape,
                                   dynamic=dynamic, static_target=static_target)

    def _setup_parameters(self):
        """
        Adds an IndepVarComp if necessary and issues appropriate connections based
        on transcription.
        """
        if self.parameter_options:
            param_comp = ParameterComp()
            self.add_subsystem('param_comp', subsys=param_comp, promotes_inputs=['*'], promotes_outputs=['*'])

            for name, options in self.parameter_options.items():
                tgts = options['targets']

                if tgts is None:
                    # The user is implicitly connecting to input parameters in all phases.
                    # No need to create input parameters in each phase.
                    continue

                for phase_name, phs in self._phases.items():
                    if phase_name not in tgts or isinstance(tgts[phase_name], str):
                        # If user omitted this phase from targets, we will try to connect
                        # to an existing input parameter in the phase.
                        # If the target for this phase is a string, assume the user specified the
                        # name of an input parameter in the phase for this parameter.
                        # Skip addition of input parameter to this phase.
                        continue
                    elif tgts[phase_name] is None:
                        # Targets for this phase are explicitly None.
                        # Skip addition of input parameter to this phase.
                        continue
                    elif isinstance(tgts[phase_name], Sequence):
                        # User specified ODE targets for this parameter in this phase.
                        # We need to add an input parameter to this phase.

                        # The default target in the phase is name unless otherwise specified.
                        kwargs = {'static_target': options['static_target'],
                                  'units': options['units'],
                                  'val': options['val'],
                                  'shape': options['shape'],
                                  'ref0': options['ref0'],
                                  'ref': options['ref'],
                                  'adder': options['adder'],
                                  'scaler': options['scaler'],
                                  'opt': options['opt'],
                                  'lower': options['lower'],
                                  'upper': options['upper'],
                                  'targets': tgts[phase_name]}

                        if not self.options['sim_mode']:
                            phs.add_parameter(name, **kwargs)

    def _setup_linkages(self):
        has_linkage_constraints = False

        err_template = '{traj}: Phase `{phase1}` links variable `{var1}` to phase ' \
                       '`{phase2}` state variable `{var2}` by connection, but phase `{phase2}` ' \
                       'is an AnalyticPhase and does not support linking initial state ' \
                       'values with option `connected=True`.'

        for pair, var_dict in self._linkages.items():

            for name in pair:
                if name not in self._phases:
                    raise ValueError(f'Invalid linkage.  Phase \'{name}\' does not exist in '
                                     f'trajectory \'{self.pathname}\'.')

            phase2 = self._phases[pair[1]]

            for var_pair, options in var_dict.items():
                var1, var2 = var_pair

                if options['connected']:
                    if var2 == 'time':
                        phase2.set_time_options(input_initial=True, fix_initial=False)
                    elif var2 == '*':
                        phase2.set_time_options(input_initial=True, fix_initial=False)
                        for state_name in phase2.state_options:
                            if isinstance(phase2, AnalyticPhase):
                                raise RuntimeError(err_template.format(traj=self.pathname, phase1=pair[0],
                                                                       phase2=pair[1], var1=var1, var2=var2))
                            else:
                                phase2.set_state_options(state_name, input_initial=True, fix_initial=_unspecified)
                    elif var2 in phase2.state_options:
                        if isinstance(phase2, AnalyticPhase):
                            raise RuntimeError(err_template.format(traj=self.pathname, phase1=pair[0],
                                                                   phase2=pair[1], var1=var1, var2=var2))
                        else:
                            phase2.set_state_options(var2, input_initial=True, fix_initial=False)
                    elif var2 in phase2.parameter_options:
                        phase2.set_parameter_options(var2, opt=False)
                else:
                    has_linkage_constraints = True

        if has_linkage_constraints:
            self.add_subsystem('linkages', PhaseLinkageComp())

    def setup(self):
        """
        Setup the Trajectory Group.
        """
        super(Trajectory, self).setup()

        if self.parameter_options:
            self._setup_parameters()

        # This will override the existing phases attribute with the same thing.
        self.add_subsystem('phases', subsys=self.phases)

        if self._linkages:
            self._setup_linkages()

    def _configure_parameters(self):
        """
        Configure connections from input or design parameters to the appropriate targets
        in each phase.
        """
        parameter_options = self.parameter_options
        promoted_inputs = []

        for name, options in parameter_options.items():
            promoted_inputs.append(f'parameters:{name}')
            targets = options['targets']

            # For each phase, use introspection to get the units and shape.
            # If units do not match across all phases, require user to set them.
            # If shapes do not match across all phases, this is an error.
            tgts = []
            tgt_units = {}
            tgt_shapes = {}

            for phase_name, phs in self._phases.items():

                if targets is None or phase_name not in targets:
                    # Attempt to connect to an input parameter of the same name in the phase, if
                    # it exists.
                    if name in phs.parameter_options:
                        tgt = f'{phase_name}.parameters:{name}'
                        tgt_shapes[phs.name] = phs.parameter_options[name]['shape']
                        tgt_units[phs.name] = phs.parameter_options[name]['units']
                    else:
                        continue
                elif targets[phase_name] is None:
                    # Connections to this phase are explicitly omitted
                    continue
                elif isinstance(targets[phase_name], str):
                    if targets[phase_name] in phs.parameter_options:
                        # Connect to an input parameter with a different name in this phase
                        tgt = f'{phase_name}.parameters:{targets[phase_name]}'
                        tgt_shapes[phs.name] = phs.parameter_options[targets[phase_name]]['shape']
                        tgt_units[phs.name] = phs.parameter_options[targets[phase_name]]['units']
                    else:
                        msg = f'Invalid target for trajectory `{self.pathname}` parameter `{name}` in phase ' \
                              f"`{phase_name}`.\nTarget for phase `{phase_name}` is '{targets[phase_name]}' but " \
                              f"the phase has no such parameter."
                        raise ValueError(msg)
                elif isinstance(targets[phase_name], Sequence):
                    if name in phs.parameter_options:
                        # User gave a list of ODE targets which were passed to the creation of a
                        # new input parameter in setup, just connect to that new input parameter
                        tgt = f'{phase_name}.parameters:{name}'
                        tgt_shapes[phs.name] = phs.parameter_options[name]['shape']
                        tgt_units[phs.name] = phs.parameter_options[name]['units']
                    else:
                        msg = f'Invalid target for trajectory `{self.pathname}` parameter `{name}` in phase ' \
                              f"`{phase_name}`.\nThe phase did not add the parameter as expected. Please file an " \
                              f"issue with the Dymos development team at https://github.com/OpenMDAO/dymos"
                        raise RuntimeError(msg)
                else:
                    raise ValueError(f'Unhandled target(s) ({targets[phase_name]}) for parameter {name} in '
                                     f'phase {phase_name}. If connecting to ODE inputs in the phase, '
                                     f'format the targets as a sequence of strings.')
                tgts.append(tgt)

            if not tgts:
                # Find the reason
                if targets is None:
                    reason = f'Option `targets=None` but no phase in the trajectory has a parameter named `{name}`.'
                elif all([t is None for t in targets.values()]) and targets.keys() == self._phases.keys():
                    reason = f'Option `targets` is a dictionary keyed by phase name but target for each phase is None.'
                raise ValueError(f'No target was found for trajectory parameter `{name}` in any phase.\n{reason}')

            if options['shape'] is _unspecified:
                if len(set(tgt_shapes.values())) == 1:
                    options['shape'] = next(iter(tgt_shapes.values()))
                else:
                    raise ValueError(f'Parameter {name} in Trajectory {self.pathname} is connected to '
                                     f'targets in multiple phases that have different shapes.')

            if options['units'] is _unspecified:
                tgt_units_set = set(tgt_units.values())
                if len(tgt_units_set) == 1:
                    options['units'] = tgt_units_set.pop()
                else:
                    ValueError(f'Parameter {name} in Trajectory {self.pathname} is connected to '
                               f'targets in multiple phases that have different units. You must '
                               f'explicitly provide units for the parameter since they cannot be '
                               f'inferred.')

            param_comp = self._get_subsystem('param_comp')
            param_comp.add_parameter(name, val=options['val'], shape=options['shape'], units=options['units'])
            if options['opt']:
                lb = -INF_BOUND if options['lower'] is None else options['lower']
                ub = INF_BOUND if options['upper'] is None else options['upper']

                self.add_design_var(name=f'parameters:{name}',
                                    lower=lb,
                                    upper=ub,
                                    scaler=options['scaler'],
                                    adder=options['adder'],
                                    ref0=options['ref0'],
                                    ref=options['ref'])

            self.connect(f'parameter_vals:{name}', tgts)

        return promoted_inputs

    def _configure_phase_options_dicts(self):
        """
        Called during configure if we are under MPI. Loops over all phases and broadcasts the shape
        and units options to all procs for all dymos variables.
        """
        for phase in self._phases.values():
            all_dicts = [phase.state_options, phase.control_options, phase.parameter_options,
                         phase.polynomial_control_options]

            for opt_dict in all_dicts:
                for options in opt_dict.values():

                    all_ranks = self.comm.allgather(options['shape'])
                    for item in all_ranks:
                        if item not in {None, _unspecified}:
                            options['shape'] = item
                            break
                    else:
                        raise RuntimeError('Unexpectedly found no valid shape.')

                    all_ranks = self.comm.allgather(options['units'])
                    for item in all_ranks:
                        if item is not _unspecified:
                            options['units'] = item
                            break
                    else:
                        raise RuntimeError('Unexpectedly found no valid units.')

    def _update_linkage_options_configure(self, linkage_options):
        """
        Called during configure to return the source paths, units, and shapes of variables
        in linkages.

        Parameters
        ----------
        phases : Sequence of (str, str)
            The names of the phases involved in the linkage.
        vars : Sequence of (str, str)
            The paths of the variables involved in the linkage.
        options : dict
            The linkage options set during `add_linkage_constraint`.

        Returns
        -------

        """
        phase_name_a = linkage_options['phase_a']
        phase_name_b = linkage_options['phase_b']
        var_a = linkage_options['var_a']
        var_b = linkage_options['var_b']

        info_str = f'Error in linking {var_a} from {phase_name_a} to {var_b} in {phase_name_b}'

        phase_a = self._get_subsystem(f'phases.{phase_name_a}')
        phase_b = self._get_subsystem(f'phases.{phase_name_b}')

        phases = {'a': phase_a, 'b': phase_b}

        classes = {'a': phase_a.classify_var(var_a),
                   'b': phase_b.classify_var(var_b)}

        sources = {'a': None, 'b': None}
        vars = {'a': var_a, 'b': var_b}
        units = {'a': _unspecified, 'b': _unspecified}
        shapes = {'a': _unspecified, 'b': _unspecified}

        for i in ('a', 'b'):
            if classes[i] == 'time':
                sources[i] = 'timeseries.time'
                shapes[i] = (1,)
                units[i] = phases[i].time_options['units']
            elif classes[i] == 'time_phase':
                sources[i] = 'timeseries.time_phase'
                units[i] = phases[i].time_options['units']
                shapes[i] = (1,)
            elif classes[i] == 'state':
                sources[i] = f'timeseries.states:{vars[i]}'
                units[i] = phases[i].state_options[vars[i]]['units']
                shapes[i] = phases[i].state_options[vars[i]]['shape']
            elif classes[i] in {'indep_control', 'input_control'}:
                sources[i] = f'timeseries.controls:{vars[i]}'
                units[i] = phases[i].control_options[vars[i]]['units']
                shapes[i] = phases[i].control_options[vars[i]]['shape']
            elif classes[i] in {'control_rate', 'control_rate2'}:
                sources[i] = f'timeseries.control_rates:{vars[i]}'
                control_name = vars[i][:-5] if classes[i] == 'control_rate' else vars[i][:-6]
                units[i] = phases[i].control_options[control_name]['units']
                deriv = 1 if classes[i].endswith('rate') else 2
                units[i] = get_rate_units(units[i], phases[i].time_options['units'], deriv=deriv)
                shapes[i] = phases[i].control_options[control_name]['shape']
            elif classes[i] in {'indep_polynomial_control', 'input_polynomial_control'}:
                sources[i] = f'timeseries.polynomial_controls:{vars[i]}'
                units[i] = phases[i].polynomial_control_options[vars[i]]['units']
                shapes[i] = phases[i].polynomial_control_options[vars[i]]['shape']
            elif classes[i] in {'polynomial_control_rate', 'polynomial_control_rate2'}:
                sources[i] = f'timeseries.polynomial_control_rates:{vars[i]}'
                control_name = vars[i][:-5] if classes[i] == 'polynomial_control_rate' else vars[i][:-6]
                control_units = phases[i].polynomial_control_options[control_name]['units']
                time_units = phases[i].time_options['units']
                deriv = 1 if classes[i].endswith('rate') else 2
                units[i] = get_rate_units(control_units, time_units, deriv=deriv)
                shapes[i] = phases[i].polynomial_control_options[control_name]['shape']
            elif classes[i] == 'parameter':
                sources[i] = f'parameter_vals:{vars[i]}'
                units[i] = phases[i].parameter_options[vars[i]]['units']
                shapes[i] = phases[i].parameter_options[vars[i]]['shape']
            else:
                rhs_source = phases[i].options['transcription']._rhs_source
                sources[i] = f'{rhs_source}.{vars[i]}'
                try:
                    meta = get_source_metadata(phases[i]._get_subsystem(rhs_source), vars[i], user_units=units[i],
                                               user_shape=_unspecified)
                    shapes[i] = meta['shape']
                    units[i] = meta['units']
                except ValueError as e:
                    raise RuntimeError(f'{info_str}: Unable to find variable \'{vars[i]}\' in '
                                       f'phase \'{phases[i].pathname}\' or its ODE.')

        linkage_options._src_a = sources['a']
        linkage_options._src_b = sources['b']
        linkage_options['shape'] = shapes['b']

        if linkage_options['units_a'] is _unspecified:
            linkage_options['units_a'] = units['a']

        if linkage_options['units_b'] is _unspecified:
            linkage_options['units_b'] = units['b']

        if (linkage_options['units_a'] != linkage_options['units_b']) and \
                linkage_options['units'] is _unspecified:
            raise ValueError(f'{info_str}: Linkage units were not specified but the units of '
                             f'var_a ({units["a"]}) and var_b ({units["b"]}) are not the same. '
                             f'Units for this linkage constraint must be specified explicitly.')
        else:
            linkage_options['units'] = units['a']

    def _expand_star_linkage_configure(self):
        """
        Finds the variable pair ('*', '*') and expands it out to time and all states if found.

        Returns
        -------
        dict
            The updated dictionary of linkages with '*' expanded to match time and all states at
            a phase boundary.

        """
        linkages_copy = deepcopy(self._linkages)
        for phase_pair, var_dict in linkages_copy.items():
            phase_name_a, phase_name_b = phase_pair

            phase_b = self._get_subsystem(f'phases.{phase_name_b}')

            for var_pair in var_dict.keys():
                if tuple(var_pair) == ('*', '*'):
                    options = var_dict[var_pair]
                    self.add_linkage_constraint(phase_name_a, phase_name_b, var_a='time',
                                                var_b='time', loc_a=options['loc_a'],
                                                loc_b=options['loc_b'], sign_a=options['sign_a'],
                                                sign_b=options['sign_b'])
                    for state_name in phase_b.state_options:
                        self.add_linkage_constraint(phase_name_a, phase_name_b, var_a=state_name,
                                                    var_b=state_name, loc_a=options['loc_a'],
                                                    loc_b=options['loc_b'],
                                                    sign_a=options['sign_a'],
                                                    sign_b=options['sign_b'])
                    self._linkages[phase_pair].pop(var_pair)

    def _is_valid_linkage(self, phase_name_a, phase_name_b, loc_a, loc_b, var_a, var_b):
        """
        Validates linkage constraints.

        Ensures that the posed linkage constraint can be satisfied by checking that the optimizer
        has the freedom to change the linked variable value on either side of the linkage.

        This check errs on the side of permitting linkages if their validity cannot be confirmed.

        Parameters
        ----------
        phase_name_a : str
            The phase name on the first side of the linkage.
        phase_name_b : str
            The phase name on the second side of the linkage.
        loc_a : str
            The "location" of the first side of the linkage, either "initial" or "final".
        loc_b : str
            The "location" of the second side of the linkage, either "initial" or "final".
        var_a : str
            The variable name of the first side of the linkage.
        var_b : str
            The variable name of the second side of the linkage.

        Returns
        -------
        bool
            True if the linkage constraint is valid.
        str
            A message explaining why the linkage is not valid.  Empty for valid linkages.
        """
        phase_a = self._get_subsystem(f'phases.{phase_name_a}')
        phase_b = self._get_subsystem(f'phases.{phase_name_b}')

        var_cls_a = phase_a.classify_var(var_a)
        var_cls_b = phase_b.classify_var(var_b)

        if var_cls_a == 'time':
            var_a_fixed = phase_a.is_time_fixed(loc_a)
        elif var_cls_a == 'state':
            var_a_fixed = phase_a.is_state_fixed(var_a, loc_a)
        else:
            var_a_fixed = False

        if var_cls_b == 'time':
            var_b_fixed = phase_b.is_time_fixed(loc_b)
        elif var_cls_b == 'state':
            var_b_fixed = phase_b.is_state_fixed(var_b, loc_b)
        else:
            var_b_fixed = False

        if var_a_fixed and var_b_fixed:
            return False, f'Cannot link {loc_a} value of "{var_a}" in {phase_name_a} to {loc_b} ' \
                          f'value of "{var_b}" in {phase_name_b}.  Values on both sides of the linkage ' \
                          'are fixed.'
        else:
            return True, ''

    def _configure_linkages(self):
        connected_linkage_inputs = []

        def _print_on_rank(rank=0, *args, **kwargs):
            if self.comm.rank == rank:
                print(*args, **kwargs)

        def _get_prefixed_var(var, phase):
            class_var = phase.classify_var(var)
            prefixes = {'time': '',
                        'time_phase': '',
                        'state': 'states:',
                        'parameter': 'parameters:',
                        'input_control': 'controls:',
                        'indep_control': 'controls:',
                        'control_rate': 'control_rates:',
                        'control_rate2': 'control_rates:',
                        'input_polynomial_control': 'polynomial_controls:',
                        'indep_polynomial_control': 'polynomial_controls:',
                        'polynomial_control_rate': 'polynomial_control_rates:',
                        'polynomial_control_rate2': 'polynomial_control_rates:',
                        'ode': ''
                        }
            return f'{prefixes[class_var]}{var}'

        # First, if the user requested all states and time be continuous ('*', '*'), then
        # expand it out.
        self._expand_star_linkage_configure()

        _print_on_rank(f'--- Linkage Report [{self.pathname}] ---')

        indent = '    '

        linkage_comp = self._get_subsystem('linkages')

        for phase_pair, var_dict in self._linkages.items():
            phase_name_a, phase_name_b = phase_pair
            _print_on_rank(f'{indent}--- {phase_name_a} - {phase_name_b} ---')

            phase_a = self._get_subsystem(f'phases.{phase_name_a}')
            phase_b = self._get_subsystem(f'phases.{phase_name_b}')

            # Pull out the maximum variable name length of all variables to make the print nicer.
            var_len_a = [len(_get_prefixed_var(var_pair[0], phase_a)) for var_pair in var_dict]
            var_len_b = [len(_get_prefixed_var(var_pair[1], phase_b)) for var_pair in var_dict]
            padding_a = max(var_len_a) + 2
            padding_b = max(var_len_b) + 2

            for var_pair, options in var_dict.items():
                var_a, var_b = var_pair
                loc_a = options['loc_a']
                loc_b = options['loc_b']

                class_a = phase_a.classify_var(var_a)
                class_b = phase_b.classify_var(var_b)

                self._update_linkage_options_configure(options)

                src_a = options._src_a
                src_b = options._src_b

                if class_a == 'time':
                    fixed_a = phase_a.is_time_fixed(loc_a)
                elif class_a == 'state':
                    fixed_a = phase_a.is_state_fixed(var_a, loc_a)
                elif class_a in {'input_control', 'indep_control'}:
                    fixed_a = phase_a.is_control_fixed(var_a, loc_a)
                elif class_a in {'input_polynomial_control', 'indep_polynomial_control'}:
                    fixed_a = phase_a.is_polynomial_control_fixed(var_a, loc_a)
                elif class_a == 'parameter':
                    fixed_a = not phase_a.parameter_options[var_a]['opt']
                else:
                    fixed_a = True

                if class_b == 'time':
                    fixed_b = phase_b.is_time_fixed(loc_b)
                elif class_b == 'state':
                    fixed_b = phase_b.is_state_fixed(var_b, loc_b)
                elif class_b in {'input_control', 'indep_control'}:
                    fixed_b = phase_b.is_control_fixed(var_b, loc_b)
                elif class_b in {'input_polynomial_control', 'indep_polynomial_control'}:
                    fixed_b = phase_b.is_polynomial_control_fixed(var_b, loc_b)
                elif class_a == 'parameter':
                    fixed_a = not phase_b.parameter_options[var_a]['opt']
                else:
                    fixed_b = True

                prefixed_a = _get_prefixed_var(var_a, phase_a)
                prefixed_b = _get_prefixed_var(var_b, phase_b)

                str_fixed_a = '*' if fixed_a else ''
                str_fixed_b = '*' if fixed_b else ''

                if options['connected']:
                    if class_b == 'time':
                        self.connect(f'{phase_name_a}.{src_a}',
                                     f'{phase_name_b}.t_initial',
                                     src_indices=[-1], flat_src_indices=True)
                    elif class_b == 'state':
                        tgt_b = f'initial_states:{var_b}'
                        self.connect(f'{phase_name_a}.{src_a}',
                                     f'{phase_name_b}.{tgt_b}',
                                     src_indices=om.slicer[-1, ...])
                    elif class_b == 'parameter':
                        tgt_b = f'parameters:{var_b}'
                        self.connect(f'{phase_name_a}.{src_a}',
                                     f'{phase_name_b}.{tgt_b}',
                                     src_indices=om.slicer[-1, ...])
                    else:
                        msg = f'Could not create connection linkage from phase `{phase_name_a}` ' \
                              f'variable `{var_a}` to phase `{phase_name_b}` variable `{var_b}`. ' \
                              f'For direct connections, the target variable must be `time`, a ' \
                              f'state, or a parameter in the phase.\nEither remove the linkage or specify ' \
                              f'`connected=False` to enforce it via an optimization constraint.'
                        raise om.OpenMDAOWarning(msg)
                    _print_on_rank(f'{indent * 2}{prefixed_a:<{padding_a}s} [{loc_a}{str_fixed_a}] ->  '
                                   f'{prefixed_b:<{padding_b}s} [{loc_b}{str_fixed_b}]')
                else:
                    is_valid, msg = self._is_valid_linkage(phase_name_a, phase_name_b,
                                                           loc_a, loc_b, var_a, var_b)

                    if not is_valid:
                        raise ValueError(f'Invalid linkage in Trajectory {self.pathname}: {msg}')

                    linkage_comp.add_linkage_configure(options)

                    if options._input_a not in connected_linkage_inputs:
                        self.connect(f'{phase_name_a}.{src_a}',
                                     f'linkages.{options._input_a}',
                                     src_indices=om.slicer[[0, -1], ...])
                        connected_linkage_inputs.append(options._input_a)

                    if options._input_b not in connected_linkage_inputs:
                        self.connect(f'{phase_name_b}.{src_b}',
                                     f'linkages.{options._input_b}',
                                     src_indices=om.slicer[[0, -1], ...])
                        connected_linkage_inputs.append(options._input_b)

                    _print_on_rank(f'{indent * 2}{prefixed_a:<{padding_a}s} [{loc_a}{str_fixed_a}] ==  '
                                   f'{prefixed_b:<{padding_b}s} [{loc_b}{str_fixed_b}]')

        _print_on_rank('\n* : Value is fixed or is an input.\n')

    def configure(self):
        """
        Configure the Trajectory Group.

        This method is used to handle connections to the phases in the Trajectory, since
        setup has already been called on all children of the Trajectory, we can query them for
        variables at this point.
        """
        if self.parameter_options:
            self._configure_parameters()

        if self._linkages:
            if MPI:
                self._configure_phase_options_dicts()
            self._configure_linkages()

        self._constraint_report(outstream=sys.stdout)

        # promote everything else out of phases
        self.promotes('phases', inputs=['*'], outputs=['*'])

    def add_linkage_constraint(self, phase_a, phase_b, var_a, var_b, loc_a='final', loc_b='initial',
                               sign_a=1.0, sign_b=-1.0, units=_unspecified, lower=None, upper=None,
                               equals=None, scaler=None, adder=None, ref0=None, ref=None,
                               linear=False, connected=False):
        """
        Explicitly add a single phase linkage constraint.

        Phase linkage constraints are enforced by constraining the following equation:

        sign_a * var_a + sign_b * var_b

        The resulting value of this equation is constrained.  This can satisfy 'coupling' or
        'linkage' conditions across phase boundaries:  enforcing continuity,
        common initial conditions, or common final conditions.

        With default values, this equation can be used to enforce variable continuity at phase
        boundaries.  For instance, constraining some variable `x` (either a state, control,
        parameter, or output of the ODE) to have the same value at the final point of phase 'foo'
        and the initial point of phase 'bar' is accomplished by:

        ```
        add_linkage_constraint('foo', 'bar', 'x', 'x')
        ```

        We may sometimes want two phases to have the same value of some variable at the start of
        each phase:

        ```
        add_linkage_constraint('foo', 'bar', 'x', 'x', loc_a='initial', loc_b='initial')
        ```

        (Here the specification of loc_b is unnecessary but helps in the clarity of whats going on.)

        Or perhaps a phase has cyclic behavior.  We may not know the exact value of some variable
        `x` at the start and end of the phase `foo`, but it must be the same value at each point.

        ```
        add_linkage_constraint('foo', 'foo', 'x', 'x')
        ```

        If `lower`, `upper`, and `equals` are all `None`, then dymos will use `equals=0` by default.
        If the continuity condition is limited by some bounds instead, lower and upper can be used.
        For instance, perhaps the velocity ('vel') is allowed to have an impulsive change within
        a certain magnitude between two phases:

        ```
        add_linkage_constraint('foo', 'bar', 'vel', 'vel', lower=-100, upper=100, units='m/s')
        ```

        Parameters
        ----------
        phase_a : str
            The first phase in the linkage constraint.
        phase_b : str
            The second phase in the linkage constraint.
        var_a : str
            The linked variable from the first phase in the linkage constraint.
        var_b : str
            The linked variable from the second phase in the linkage constraint.
        loc_a : str
            The location of the variable in the first phase of the linkage constraint (one of
            'initial' or 'final').
        loc_b : str
            The location of the variable in the second phase of the linkage constraint (one of
            'initial' or 'final').
        sign_a : float
            The sign applied to the variable from the first phase in the linkage constraint.
        sign_b : float
            The sign applied to the variable from the second phase in the linkage constraint.
        units : str or None or _unspecified
            Units of the linkage.  If _unspecified, dymos will use the units from the variable
            in the first phase of the linkage.  Units of the two specified variables must be
            compatible.
        lower : float or array or None
            The lower bound applied as a constraint on the linkage equation.
        upper : float or array or None
            The upper bound applied as a constraint on the linkage equation.
        equals : float or array or None
            Specifies a targeted value for an equality constraint on the linkage equation.
        scaler : float or array or None
            The scaler of the linkage constraint.
        adder : float or array or None
            The adder of the linkage constraint.
        ref0 : float or array or None
            The zero-reference value of the linkage constraint.
        ref : float or array or None
            The unit-reference value of the linkage constraint.
        linear : bool
            If True, treat this variable as a linear constraint, otherwise False.  Linear
            constraints should only be applied if the variable on each end of the linkage is a
            design variable or a linear function of one.
        connected : bool
            If True, this constraint is enforced by direct connection rather than a constraint
            for the optimizer.  This is only valid for states and time.
        """
        if connected:
            invalid_options = []
            for arg in ['lower', 'upper', 'equals', 'scaler', 'adder', 'ref0', 'ref']:
                if locals()[arg] is not None:
                    invalid_options.append(arg)
            if locals()['linear']:
                invalid_options.append('linear')
            if invalid_options:
                msg = f'Invalid option in linkage between {phase_a}:{var_a} and {phase_b}:{var_b} ' \
                      f'in trajectory {self.pathname}. The following options for ' \
                      f'add_linkage_constraint were specified but not valid when ' \
                      f'option \'connected\' is True: ' + ' '.join(invalid_options)
                warnings.warn(msg)

        if (phase_a, phase_b) not in self._linkages:
            self._linkages[phase_a, phase_b] = OrderedDict()

        self._linkages[phase_a, phase_b][var_a, var_b] = d = LinkageOptionsDictionary()
        d['phase_a'] = phase_a
        d['phase_b'] = phase_b
        d['var_a'] = var_a
        d['var_b'] = var_b
        d['loc_a'] = loc_a
        d['loc_b'] = loc_b
        d['sign_a'] = sign_a
        d['sign_b'] = sign_b
        d['units'] = units
        d['lower'] = lower
        d['upper'] = upper
        d['equals'] = equals
        d['scaler'] = scaler
        d['adder'] = adder
        d['ref0'] = ref0
        d['ref'] = ref
        d['linear'] = linear
        d['connected'] = connected

    def link_phases(self, phases, vars=None, locs=('final', 'initial'), connected=False):
        """
        Specify that phases in the given sequence are to be assume continuity of the given variables.

        This method caches the phase linkages, and may be called multiple times to express more
        complex behavior (branching phases, phases only continuous in some variables, etc).

        The location at which the variables should be coupled in the two phases are provided
        with one of two strings:

        - 'final' specifies the value at the end of the phase (at time t_initial + t_duration)
        - 'initial' specifies the value at the start of the phase (at time t_initial)

        Parameters
        ----------
        phases : sequence of str
            The names of the phases in this trajectory to be sequentially linked.
        vars : sequence of str
            The variables in the phases to be linked, or '*'.  Providing '*' will time and all
            states.  Linking control values or rates requires them to be listed explicitly.
        locs : tuple of str
            A two-element tuple of a location specification.  For every pair in
            phases, the location specification refers to which location in the first phase is
            connected to which location in the second phase.  If the user wishes to specify
            different locations for different phase pairings, those phase pairings must be made
            in separate calls to link_phases.
        connected : bool
            Set to True to directly connect the phases being linked. Otherwise, create constraints
            for the optimizer to solve.

        See Also
        --------
        add_linkage_constraint : Explicitly add a single phase linkage constraint.
        """
        num_links = len(phases) - 1

        if num_links <= 0:
            raise ValueError('Phase sequence must consists of at least two phases.')
        if isinstance(locs, Sequence) and len(locs) != 2:
            raise ValueError('If given, locs must be a sequence of two-element tuples, one pair '
                             'for each phase pair in the phases sequence')

        _vars = ['*'] if vars is None else vars

        # Resolve linkage pairs from the phases sequence
        a, b = itertools.tee(phases)
        next(b, None)

        if len(locs) == 1:
            _locs = num_links * locs
        elif len(locs) == 2:
            _locs = num_links * [locs]
        elif len(locs) == num_links:
            _locs = locs
        else:
            raise ValueError('The number of location tuples, if provided, must be one less than '
                             f'the number of phases specified.  There are {num_links} phase pairs '
                             f'but {len(locs)} location tuples specified.')

        for phase_name_a, phase_name_b, loctup in zip(a, b, _locs):
            loc_a, loc_b = loctup
            for var in _vars:
                self.add_linkage_constraint(phase_a=phase_name_a, phase_b=phase_name_b,
                                            var_a=var, var_b=var, loc_a=loc_a, loc_b=loc_b,
                                            connected=connected)

    def _constraint_report(self, outstream=sys.stdout):
        if self.options['sim_mode']:
            return

        if self.comm.rank == 0:
            printer = print
        else:
            def printer(*args, **kwargs):
                pass

        float_fmt = '6.4e'
        printer(f'\n--- Constraint Report [{self.pathname}] ---')
        indent = '    '

        def _print_constraints(phase, outstream):
            tx = phase.options['transcription']

            ode_outputs = get_promoted_vars(tx._get_ode(phase), 'output')

            ds = {'initial': phase._initial_boundary_constraints,
                  'final': phase._final_boundary_constraints,
                  'path': phase._path_constraints}

            if not (
                    phase._initial_boundary_constraints or phase._final_boundary_constraints or phase._path_constraints):
                printer(f'{2 * indent}None', file=outstream)

            for loc, d in ds.items():
                str_loc = f'[{loc}]'
                for options in d:
                    expr = options['name']
                    _, shape, units, linear = tx._get_objective_src(expr, loc, phase, ode_outputs=ode_outputs)

                    equals = options['equals']
                    lower = options['lower']
                    upper = options['upper']

                    if options['units']:
                        str_units = options['units']
                    elif units is not None:
                        str_units = units
                    else:
                        str_units = 'None'

                    if equals is not None and np.prod(np.asarray(equals).shape) != 1:
                        str_equals = f'array<{"x".join([str(i) for i in np.asarray(equals).shape])}>'
                    elif equals is not None:
                        str_equals = f'{equals:{float_fmt}}'

                    if lower is not None and np.prod(np.asarray(lower).shape) != 1:
                        str_lower = f'array<{"x".join([str(i) for i in np.asarray(lower).shape])}> <='
                    elif lower is not None:
                        str_lower = f'{lower:{float_fmt}} <='
                    else:
                        str_lower = 12 * ''

                    if upper is not None and np.prod(np.asarray(upper).shape) != 1:
                        str_upper = f'<= array<{"x".join([str(i) for i in np.asarray(upper).shape])}> '
                    elif upper is not None:
                        str_upper = f'<= {upper:{float_fmt}} '
                    else:
                        str_upper = ''

                    if equals is not None:
                        printer(f'{2 * indent}{str_loc:<10s}{str_equals} == {expr} [{str_units}]',
                                file=outstream)
                    else:
                        printer(
                            f'{2 * indent}{str_loc:<10s}{str_lower} {expr} {str_upper} [{str_units}]',
                            file=outstream)

        for phase_name, phs in self._phases.items():
            printer(f'{indent}--- {phase_name} ---', file=outstream)
            if phs._is_local:
                _print_constraints(phs, outstream)

        printer('', file=outstream)

    def simulate(self, times_per_seg=10, method=_unspecified, atol=_unspecified, rtol=_unspecified,
                 first_step=_unspecified, max_step=_unspecified, record_file=None, case_prefix=None,
                 reset_iter_counts=True, reports=False):
        """
        Simulate the Trajectory using scipy.integrate.solve_ivp.

        Parameters
        ----------
        times_per_seg : int or None
            Number of equally spaced times per segment at which output is requested.  If None,
            output will be provided at all Nodes.
        method : str
            The scipy.integrate.solve_ivp integration method.
        atol : float
            Absolute convergence tolerance for scipy.integrate.solve_ivp.
        rtol : float
            Relative convergence tolerance for scipy.integrate.solve_ivp.
        first_step : float
            Initial step size for the integration.
        max_step : float
            Maximum step size for the integration.
        record_file : str or None
            If a string, the file to which the result of the simulation will be saved.
            If None, no record of the simulation will be saved.
        case_prefix : str or None
            Prefix to prepend to coordinates when recording.
        reset_iter_counts : bool
            If True and model has been run previously, reset all iteration counters.
        reports : bool or None or str or Sequence
            Reports setting for the subproblems run under simualate.

        Returns
        -------
        problem
            An OpenMDAO Problem in which the simulation is implemented.  This Problem interface
            can be interrogated to obtain timeseries outputs in the same manner as other Phases
            to obtain results at the requested times.
        """
        sim_traj = Trajectory(sim_mode=True)

        for name, phs in self._phases.items():
            if phs.simulate_options is None:
                continue

            sim_phs = phs.get_simulation_phase(times_per_seg=times_per_seg, method=method,
                                               atol=atol, rtol=rtol, first_step=first_step,
                                               max_step=max_step, reports=reports)
            sim_traj.add_phase(name, sim_phs)

        if not sim_traj._phases:
            raise RuntimeError(f'Trajectory `{self.pathname}` has no phases that support simulation.')

        sim_traj.parameter_options.update(self.parameter_options)

        sim_prob = om.Problem(model=om.Group(), reports=reports)

        traj_name = self.name if self.name else 'sim_traj'
        sim_prob.model.add_subsystem(traj_name, sim_traj)

        if record_file is not None:
            rec = om.SqliteRecorder(record_file)
            sim_prob.add_recorder(rec)
            # record_inputs is needed to capture potential input parameters that aren't connected
            sim_prob.recording_options['record_inputs'] = True
            # record_outputs is need to capture the timeseries outputs
            sim_prob.recording_options['record_outputs'] = True

        sim_prob.setup()

        # Assign trajectory parameter values
        for name in self.parameter_options:
            sim_prob_prom_path = f'{traj_name}.parameters:{name}'
            sim_prob.set_val(sim_prob_prom_path, self.get_val(f'parameters:{name}'))

        for phase_name, phs in sim_traj._phases.items():
            phs.initialize_values_from_phase(sim_prob, self._phases[phase_name],
                                             phase_path=traj_name)

        print(f'\nSimulating trajectory {self.pathname}')
        sim_prob.run_model(case_prefix=case_prefix, reset_iter_counts=reset_iter_counts)
        print(f'Done simulating trajectory {self.pathname}')
        if record_file:
            _case_prefix = '' if case_prefix is None else f'{case_prefix}_'
            sim_prob.record(f'{_case_prefix}final')
        sim_prob.cleanup()

        return sim_prob



"""
SimulationPhase is an instance that resembles a Phase in structure but is intended for
use with scipy.solve_ivp to verify the accuracy of the implicit solutions of Dymos.
"""
import numpy as np

from scipy.integrate import solve_ivp

import openmdao.api as om
from ....utils.interpolate import LagrangeBarycentricInterpolant
from ....utils.lgl import lgl
from .ode_integration_interface import ODEIntegrationInterface
from ....phase.options import TimeOptionsDictionary, SimulateOptionsDictionary


class SegmentSimulationComp(om.ExplicitComponent):
    """
    Class definition for SegmentSimulationComp.

    SegmentSimulationComp is a component which, given values for time, states, and controls
    within a given segment, explicitly simulates the segment using scipy.integrate.solve_ivp.

    The resulting states are captured at all nodes within the segment.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare('index', desc='the index of this segment in the parent phase.')

        self.options.declare('grid_data', desc='the grid data of the corresponding phase.')

        self.options.declare('simulate_options', types=SimulateOptionsDictionary,
                             desc='The simulation options of the parent phase.')

        self.options.declare('ode_class', recordable=False,
                             desc='System defining the ODE')

        self.options.declare('ode_init_kwargs', types=dict, default={},
                             desc='Keyword arguments provided when initializing the ODE System')

        self.options.declare('time_options', types=TimeOptionsDictionary,
                             desc='Time options for the phase')

        self.options.declare('state_options', types=dict,
                             desc='Dictionary of state names/options for the segments parent Phase')

        self.options.declare('control_options', default=None, types=dict, allow_none=True,
                             desc='Dictionary of control names/options for the segments parent Phase.')

        self.options.declare('polynomial_control_options', default=None, types=dict, allow_none=True,
                             desc='Dictionary of polynomial control names/options for the segments '
                                  'parent Phase.')

        self.options.declare('parameter_options', default=None, types=dict, allow_none=True,
                             desc='Dictionary of parameter names/options for the segments '
                                  'parent Phase.')

        self.options.declare('ode_integration_interface', default=None, allow_none=True,
                             types=ODEIntegrationInterface, recordable=False,
                             desc='The instance of the ODE integration interface used to provide '
                                  'the ODE to scipy.integrate.solve_ivp in the segment.  If None,'
                                  ' a new one will be instantiated for this segment.')

        self.options.declare('output_nodes_per_seg', default=None, types=(int,), allow_none=True,
                             desc='If None, results are provided at the all nodes within each'
                                  'segment.  If an int (n) then results are provided at n '
                                  'equally distributed points in time within each segment.')

        self.options.declare('reports', default=False, desc='Reports setting for the subproblem.')

        self.recording_options['options_excludes'] = ['ode_integration_interface']

    def configure_io(self):
        """
        I/O creation is delayed until configure so we can determine variable shape and units.
        """
        idx = self.options['index']
        gd = self.options['grid_data']

        if self.options['output_nodes_per_seg'] is None:
            nnps_i = gd.subset_num_nodes_per_segment['all'][idx]
        else:
            nnps_i = self.options['output_nodes_per_seg']

        # Number of control discretization nodes per segment
        ncdsps = gd.subset_num_nodes_per_segment['control_disc'][idx]

        # Indices of the control disc nodes belonging to the current segment
        control_disc_seg_idxs = gd.subset_segment_indices['control_disc'][idx]

        # Segment tau values for the control disc nodes in the phase
        control_disc_stau = gd.node_stau[gd.subset_node_indices['control_disc']]

        # Segment tau values for the control disc nodes in the current segment
        control_disc_seg_stau = control_disc_stau[control_disc_seg_idxs[0]:control_disc_seg_idxs[1]]

        if self.options['ode_integration_interface'] is None:
            self.options['ode_integration_interface'] = ODEIntegrationInterface(
                ode_class=self.options['ode_class'],
                time_options=self.options['time_options'],
                state_options=self.options['state_options'],
                control_options=self.options['control_options'],
                polynomial_control_options=self.options['polynomial_control_options'],
                parameter_options=self.options['parameter_options'],
                ode_init_kwargs=self.options['ode_init_kwargs'],
                reports=self.options['reports'])

        self.add_input(name='time', val=np.ones(nnps_i),
                       units=self.options['time_options']['units'],
                       desc='Time at all nodes within the segment.')

        self.add_input(name='time_phase', val=np.ones(nnps_i),
                       units=self.options['time_options']['units'],
                       desc='Phase elapsed time at all nodes within the segment.')

        self.add_input(name='t_initial', val=0.0, units=self.options['time_options']['units'],
                       desc='Initial time value in the phase.')

        self.add_input(name='t_duration', val=1.0, units=self.options['time_options']['units'],
                       desc='Total time duration of the phase.')

        # Setup the initial state vector for integration
        self.state_vec_size = 0
        for name, options in self.options['state_options'].items():
            self.state_vec_size += np.prod(options['shape'])
            self.add_input(name='initial_states:{0}'.format(name), val=np.ones((1,) + options['shape']),
                           units=options['units'], desc='initial values of state {0} '
                                                        'in the segment'.format(name))
            self.add_output(name='states:{0}'.format(name),
                            val=np.ones((nnps_i,) + options['shape']),
                            units=options['units'],
                            desc='Values of state {0} at all nodes in the segment.'.format(name))

        self.initial_state_vec = np.zeros(self.state_vec_size)

        self.options['ode_integration_interface'].prob.setup(check=False)

        # Setup the control interpolants
        if self.options['control_options']:
            for name, options in self.options['control_options'].items():
                self.add_input(name='controls:{0}'.format(name),
                               val=np.ones(((ncdsps,) + options['shape'])),
                               units=options['units'],
                               desc='Values of control {0} at control discretization '
                                    'nodes within the segment.'.format(name))
                interp = LagrangeBarycentricInterpolant(control_disc_seg_stau, options['shape'])
                self.options['ode_integration_interface'].set_interpolant(name, interp)

        if self.options['polynomial_control_options']:
            for name, options in self.options['polynomial_control_options'].items():
                poly_control_disc_ptau, _ = lgl(options['order'] + 1)
                self.add_input(name='polynomial_controls:{0}'.format(name),
                               val=np.ones(((options['order'] + 1,) + options['shape'])),
                               units=options['units'],
                               desc='Values of polynomial control {0} at control discretization '
                                    'nodes within the phase.'.format(name))
                interp = LagrangeBarycentricInterpolant(poly_control_disc_ptau, options['shape'])
                self.options['ode_integration_interface'].set_interpolant(name, interp)

        self.declare_partials(of='*', wrt='*', method='fd')

    def compute(self, inputs, outputs):
        """
        Compute component outputs.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        idx = self.options['index']
        gd = self.options['grid_data']
        iface_prob = self.options['ode_integration_interface'].prob

        # Create the vector of initial state values
        self.initial_state_vec[:] = 0.0
        pos = 0
        for name, options in self.options['state_options'].items():
            size = np.prod(options['shape'])
            self.initial_state_vec[pos:pos + size] = \
                np.ravel(inputs['initial_states:{0}'.format(name)])
            pos += size

        # Setup the control interpolants
        if self.options['control_options']:
            t0_seg = inputs['time'][0]
            tf_seg = inputs['time'][-1]
            for name, options in self.options['control_options'].items():
                ctrl_vals = inputs['controls:{0}'.format(name)]
                self.options['ode_integration_interface'].setup_interpolant(name,
                                                                            x0=t0_seg,
                                                                            xf=tf_seg,
                                                                            f_j=ctrl_vals)

        # Setup the polynomial control interpolants
        if self.options['polynomial_control_options']:
            t0_phase = inputs['t_initial']
            tf_phase = inputs['t_initial'] + inputs['t_duration']
            for name, options in self.options['polynomial_control_options'].items():
                ctrl_vals = inputs['polynomial_controls:{0}'.format(name)]
                self.options['ode_integration_interface'].setup_interpolant(name,
                                                                            x0=t0_phase,
                                                                            xf=tf_phase,
                                                                            f_j=ctrl_vals)

        # Set the values of t_initial and t_duration
        iface_prob.set_val('t_initial',
                           val=inputs['t_initial'],
                           units=self.options['time_options']['units'])

        iface_prob.set_val('t_duration',
                           val=inputs['t_duration'],
                           units=self.options['time_options']['units'])

        # Set the values of the phase parameters
        if self.options['parameter_options']:
            for param_name, options in self.options['parameter_options'].items():
                val = inputs['parameters:{0}'.format(param_name)]
                iface_prob.set_val('parameters:{0}'.format(param_name),
                                   val=val,
                                   units=options['units'])

        # Setup the evaluation times.
        if self.options['output_nodes_per_seg'] is None:
            # Output nodes given as subset, convert segment tau of nodes to time
            i1, i2 = gd.subset_segment_indices['all'][idx, :]
            indices = gd.subset_node_indices['all'][i1:i2]
            nodes_eval = gd.node_stau[indices]  # evaluation nodes in segment tau space
            t_initial = inputs['time'][0]
            t_duration = inputs['time'][-1] - t_initial
            t_eval = t_initial + 0.5 * (nodes_eval + 1) * t_duration
        else:
            # Output nodes given as number, linspace them across the segment
            t_eval = np.linspace(inputs['time'][0], inputs['time'][-1],
                                 self.options['output_nodes_per_seg'])

        # Perform the integration using solve_ivp
        sim_options = {key: val for key, val in self.options['simulate_options'].items()}

        sol = solve_ivp(fun=self.options['ode_integration_interface'],
                        t_span=(inputs['time'][0], inputs['time'][-1]),
                        y0=self.initial_state_vec,
                        t_eval=t_eval,
                        **sim_options)

        if not sol.success:
            raise om.AnalysisError(f'solve_ivp failed: {sol.message} Dynamics changing '
                                   f'too dramatically')

        # Extract the solution
        pos = 0
        for name, options in self.options['state_options'].items():
            size = np.prod(options['shape'])
            outputs['states:{0}'.format(name)] = sol.y[pos:pos+size, :].T
            pos += size
